"""Define border styles."""

from __future__ import annotations

from enum import Enum
from functools import lru_cache, total_ordering
from typing import TYPE_CHECKING, NamedTuple

from prompt_toolkit.cache import FastDictCache

if TYPE_CHECKING:
    pass


class GridPart(Enum):
    """Define the component characters of a grid.

    Character naming works as follows:

                ╭┈┈┈┈┈┈┈┈LEFT
                ┊ ╭┈┈┈┈┈┈MID
                ┊ ┊ ╭┈┈┈┈SPLIT
                ┊ ┊ ┊ ╭┈┈RIGHT
                ∨ ∨ ∨ v
          TOP┈> ┏ ━ ┳ ┓
          MID┈> ┃   ┃ ┃
        SPLIT┈> ┣ ━ ╋ ┫
       BOTTOM┈> ┗ ━ ┻ ┛

    """

    TOP_LEFT = 0
    TOP_MID = 1
    TOP_SPLIT = 2
    TOP_RIGHT = 3
    MID_LEFT = 4
    MID_MID = 5
    MID_SPLIT = 6
    MID_RIGHT = 7
    SPLIT_LEFT = 8
    SPLIT_MID = 9
    SPLIT_SPLIT = 10
    SPLIT_RIGHT = 11
    BOTTOM_LEFT = 12
    BOTTOM_MID = 13
    BOTTOM_SPLIT = 14
    BOTTOM_RIGHT = 15


class DirectionFlags(NamedTuple):
    """Flag which indicate the connection of a grid node."""

    north: bool = False
    east: bool = False
    south: bool = False
    west: bool = False


class Mask:
    """A mask which allows selection of a subset of a grid.

    Masks can be combined to construct more complex masks.
    """

    def __init__(self, mask: dict[GridPart, DirectionFlags]) -> None:
        """Create a new grid mask.

        Args:
            mask: A dictionary mapping grid parts to a tuple of direction flags.
                Not all :class:`GridPart`s need to be defined - any which are not
                defined are assumed to be set entirely to :cont:`False`.

        """
        self.mask = {part: mask.get(part, DirectionFlags()) for part in GridPart}

    def __add__(self, other: Mask) -> Mask:
        """Add two masks, combining direction flags for each :class:`GridPart`.

        Args:
            other: Another mask to combine with this one

        Returns:
            A new combined mask.

        """
        return Mask(
            {
                key: DirectionFlags(
                    *(self.mask[key][i] | other.mask[key][i] for i in range(4))
                )
                for key in GridPart
            }
        )


class Masks:
    """A collection of default masks."""

    top_edge = Mask(
        {
            GridPart.TOP_LEFT: DirectionFlags(False, True, False, False),
            GridPart.TOP_MID: DirectionFlags(False, True, False, True),
            GridPart.TOP_SPLIT: DirectionFlags(False, True, False, True),
            GridPart.TOP_RIGHT: DirectionFlags(False, False, False, True),
        }
    )

    middle_edge = Mask(
        {
            GridPart.SPLIT_LEFT: DirectionFlags(False, True, False, False),
            GridPart.SPLIT_MID: DirectionFlags(False, True, False, True),
            GridPart.SPLIT_SPLIT: DirectionFlags(False, True, False, True),
            GridPart.SPLIT_RIGHT: DirectionFlags(False, False, False, True),
        }
    )

    bottom_edge = Mask(
        {
            GridPart.BOTTOM_LEFT: DirectionFlags(False, True, False, False),
            GridPart.BOTTOM_MID: DirectionFlags(False, True, False, True),
            GridPart.BOTTOM_SPLIT: DirectionFlags(False, True, False, True),
            GridPart.BOTTOM_RIGHT: DirectionFlags(False, False, False, True),
        }
    )

    left_edge = Mask(
        {
            GridPart.TOP_LEFT: DirectionFlags(False, False, True, False),
            GridPart.MID_LEFT: DirectionFlags(True, False, True, False),
            GridPart.SPLIT_LEFT: DirectionFlags(True, False, True, False),
            GridPart.BOTTOM_LEFT: DirectionFlags(True, False, False, False),
        }
    )

    center_edge = Mask(
        {
            GridPart.TOP_SPLIT: DirectionFlags(False, False, True, False),
            GridPart.MID_SPLIT: DirectionFlags(True, False, True, False),
            GridPart.SPLIT_SPLIT: DirectionFlags(True, False, True, False),
            GridPart.BOTTOM_SPLIT: DirectionFlags(True, False, False, False),
        }
    )

    right_edge = Mask(
        {
            GridPart.TOP_RIGHT: DirectionFlags(False, False, True, False),
            GridPart.MID_RIGHT: DirectionFlags(True, False, True, False),
            GridPart.SPLIT_RIGHT: DirectionFlags(True, False, True, False),
            GridPart.BOTTOM_RIGHT: DirectionFlags(True, False, False, False),
        }
    )

    inner = center_edge + middle_edge
    outer = top_edge + right_edge + bottom_edge + left_edge

    grid = inner + outer

    corners = Mask(
        {
            GridPart.TOP_LEFT: DirectionFlags(False, True, True, False),
            GridPart.TOP_RIGHT: DirectionFlags(False, False, True, True),
            GridPart.BOTTOM_LEFT: DirectionFlags(True, True, False, False),
            GridPart.BOTTOM_RIGHT: DirectionFlags(True, False, False, True),
        }
    )


def _lint_to_grid(line: LineStyle, mask: Mask) -> GridStyle:
    """Get a grid from a line and a mask."""
    return GridStyle(line, mask)


@total_ordering
class LineStyle:
    """Define a line style which can be used to draw grids.

    :class:`GridStyle`s can be created from a :class:`LineStyle` by accessing an
    attribute with the name of a default mask from :class:`Masks`.
    """

    _grid_cache: FastDictCache[tuple[LineStyle, Mask], GridStyle] = FastDictCache(
        get_value=_lint_to_grid
    )

    def __init__(
        self,
        name: str,
        rank: tuple[int, int],
        parent: LineStyle | None = None,
        visible: bool = True,
    ) -> None:
        """Create a new :class:`LineStyle`.

        Args:
            name: The name of the line style
            rank: A ranking value - this is used when two adjoining cells in a table
                have differing widths or styles to determine which should take precedence.
                The style with the higher rank is used. The rank consists of a tuple
                representing width and fanciness.
            parent: The :class:`LineStyle` from which this style should inherit if any
                characters are undefined.
            visible: A flag to indicate if the line is blank

        """
        self.name = name
        self.rank = rank
        self.children: dict[str, LineStyle] = {}
        self.parent = parent
        self.visible = visible
        if parent:
            parent.children[name] = self

    def __getattr__(self, value: str) -> GridStyle:
        """Define attribute access.

        Allows dynamic access to children and :class:`GridStyle` creation via attribute
        access.

        Args:
            value: The attribute name to access

        Returns:
            The accessed attribute

        Raises:
            AttributeError: Raised if there is no such attribute

        """
        if hasattr(Masks, value):
            mask = getattr(Masks, value)
            return self._grid_cache[self, mask]
        else:
            raise AttributeError(f"No such attribute `{value}`")

    def __dir__(self) -> list[str]:
        """List the public attributes."""
        return [x for x in Masks.__dict__ if not x.startswith("_")]

    def __lt__(self, other: LineStyle) -> bool:
        """Allow :class:`LineStyle`s to be sorted according to their rank."""
        if self.__class__ is other.__class__:
            return self.rank < other.rank
        elif other is None:
            return None
        return NotImplemented

    def __repr__(self) -> str:
        """Represent :class:`LineStyle` instances as a string."""
        return f"LineStyle({self.name})"


# Line Styles

# Equivalent to setting the border-stype to "none"
# - borders with a style of none have the lowest priority.
NoLine = LineStyle("None", rank=(0, 0), visible=False)
# Equivalent to setting the border-stype to "hidden"
# - takes precedence over all other conflicting borders.
InvisibleLine = LineStyle("Invisible", rank=(9999, 9999), parent=NoLine, visible=False)

AsciiLine = LineStyle("Ascii", rank=(1, 0))
ThinLine = LineStyle("Thin", rank=(1, 4), parent=AsciiLine)
RoundedLine = LineStyle("Rounded", rank=(1, 5), parent=ThinLine)
ThinQuadrupleDashedLine = LineStyle("QuadDashed", (1, 1), parent=ThinLine)
ThinTripleDashedLine = LineStyle("TripleDashed", (1, 2), parent=ThinLine)
ThinDoubleDashedLine = LineStyle("DoubleDashed", (1, 3), parent=ThinLine)

UpperRightEighthLine = LineStyle("UpperRightEighthLine", (2, 1), parent=ThinLine)
LowerLeftEighthLine = LineStyle("LowerLeftEighthLine", (2, 1), parent=ThinLine)

AsciiThickLine = LineStyle("AsciiDouble", rank=(3, 0), parent=AsciiLine)
ThickLine = LineStyle("Thick", rank=(3, 4), parent=ThinLine)
DoubleLine = LineStyle("Double", (3, 5), parent=ThickLine)
ThickQuadrupleDashedLine = LineStyle("QuadDashed", (3, 1), parent=ThickLine)
ThickTripleDashedLine = LineStyle("TripleDashed", (3, 2), parent=ThickLine)
ThickDoubleDashedLine = LineStyle("DoubleDashed", (3, 3), parent=ThickLine)

UpperRightQuarterLine = LineStyle("UpperRightQuarterLine", (4, 2), parent=ThickLine)
LowerLeftQuarterLine = LineStyle("LowerLeftQuarterLine", (4, 2), parent=ThickLine)

UpperRightHalfLine = LineStyle("UpperRightHalfLine", (5, 2), parent=ThickLine)
LowerLeftHalfLine = LineStyle("LowerLeftHalfLine", (5, 2), parent=ThickLine)
UpperRightHalfDottedLine = LineStyle(
    "UpperRightHalfDottedLine", (5, 1), parent=UpperRightHalfLine
)
LowerLeftHalfDottedLine = LineStyle(
    "LowerLeftHalfDottedLine", (5, 2), parent=LowerLeftHalfLine
)


FullLine = LineStyle("FullLine", (6, 2), parent=ThickLine)
FullDottedLine = LineStyle("FullDottedLine", (6, 1), parent=ThickLine)


class GridChar(NamedTuple):
    """Repreentation of a grid node character.

    The four compass points represent the line style joining from the given direction.
    """

    north: LineStyle
    east: LineStyle
    south: LineStyle
    west: LineStyle


# fmt: off
_GRID_CHARS = {
    # NONE
    GridChar(NoLine, NoLine, NoLine, NoLine): " ",
    # AsciiLine
    GridChar(AsciiLine, NoLine, AsciiLine, NoLine): "|",
    GridChar(NoLine, AsciiLine, NoLine, AsciiLine): "-",
    GridChar(AsciiLine, AsciiLine, NoLine, NoLine): "+",
    GridChar(NoLine, AsciiLine, AsciiLine, NoLine): "+",
    GridChar(NoLine, NoLine, AsciiLine, AsciiLine): "+",
    GridChar(AsciiLine, NoLine, NoLine, AsciiLine): "+",
    GridChar(AsciiLine, AsciiLine, AsciiLine, NoLine): "+",
    GridChar(NoLine, AsciiLine, AsciiLine, AsciiLine): "+",
    GridChar(AsciiLine, NoLine, AsciiLine, AsciiLine): "+",
    GridChar(AsciiLine, AsciiLine, NoLine, AsciiLine): "+",
    GridChar(AsciiLine, AsciiLine, AsciiLine, AsciiLine): "+",
    # AsciiThickLine
    GridChar(NoLine, AsciiThickLine, NoLine, AsciiThickLine): "=",
    # ThinLine
    GridChar(ThinLine, NoLine, NoLine, NoLine): "╵",
    GridChar(NoLine, ThinLine, NoLine, NoLine): "╶",
    GridChar(NoLine, NoLine, ThinLine, NoLine): "╷",
    GridChar(NoLine, NoLine, NoLine, ThinLine): "╴",
    GridChar(ThinLine, NoLine, ThinLine, NoLine): "│",
    GridChar(NoLine, ThinLine, NoLine, ThinLine): "─",
    GridChar(ThinLine, ThinLine, NoLine, NoLine): "└",
    GridChar(NoLine, ThinLine, ThinLine, NoLine): "┌",
    GridChar(NoLine, NoLine, ThinLine, ThinLine): "┐",
    GridChar(ThinLine, NoLine, NoLine, ThinLine): "┘",
    GridChar(ThinLine, ThinLine, ThinLine, NoLine): "├",
    GridChar(NoLine, ThinLine, ThinLine, ThinLine): "┬",
    GridChar(ThinLine, NoLine, ThinLine, ThinLine): "┤",
    GridChar(ThinLine, ThinLine, NoLine, ThinLine): "┴",
    GridChar(ThinLine, ThinLine, ThinLine, ThinLine): "┼",
    # Thin RoundedLine
    GridChar(RoundedLine, RoundedLine, NoLine, NoLine): "╰",
    GridChar(NoLine, RoundedLine, RoundedLine, NoLine): "╭",
    GridChar(NoLine, NoLine, RoundedLine, RoundedLine): "╮",
    GridChar(RoundedLine, NoLine, NoLine, RoundedLine): "╯",
    # Thin Dashes
    GridChar(NoLine, ThinDoubleDashedLine, NoLine, ThinDoubleDashedLine): "╌",
    GridChar(ThinDoubleDashedLine, NoLine, ThinDoubleDashedLine, NoLine): "╎",
    GridChar(NoLine, ThinTripleDashedLine, NoLine, ThinTripleDashedLine): "┄",
    GridChar(ThinTripleDashedLine, NoLine, ThinTripleDashedLine, NoLine): "┆",
    GridChar(NoLine, ThinQuadrupleDashedLine, NoLine, ThinQuadrupleDashedLine): "┈",
    GridChar(ThinQuadrupleDashedLine, NoLine, ThinQuadrupleDashedLine, NoLine): "┊",
    # Double
    GridChar(DoubleLine, NoLine, NoLine, NoLine): "║",
    GridChar(NoLine, DoubleLine, NoLine, NoLine): "═",
    GridChar(NoLine, NoLine, DoubleLine, NoLine): "║",
    GridChar(NoLine, NoLine, NoLine, DoubleLine): "═",
    GridChar(DoubleLine, NoLine, DoubleLine, NoLine): "║",
    GridChar(NoLine, DoubleLine, NoLine, DoubleLine): "═",
    GridChar(DoubleLine, DoubleLine, NoLine, NoLine): "╚",
    GridChar(NoLine, DoubleLine, DoubleLine, NoLine): "╔",
    GridChar(NoLine, NoLine, DoubleLine, DoubleLine): "╗",
    GridChar(DoubleLine, NoLine, NoLine, DoubleLine): "╝",
    GridChar(DoubleLine, DoubleLine, DoubleLine, NoLine): "╠",
    GridChar(NoLine, DoubleLine, DoubleLine, DoubleLine): "╦",
    GridChar(DoubleLine, NoLine, DoubleLine, DoubleLine): "╣",
    GridChar(DoubleLine, DoubleLine, NoLine, DoubleLine): "╩",
    GridChar(DoubleLine, DoubleLine, DoubleLine, DoubleLine): "╬",
    # Double / ThinLine
    GridChar(ThinLine, DoubleLine, ThinLine, DoubleLine): "╪",
    GridChar(DoubleLine, ThinLine, DoubleLine, ThinLine): "╫",
    GridChar(DoubleLine, ThinLine, NoLine, NoLine): "╙",
    GridChar(NoLine, DoubleLine, ThinLine, NoLine): "╒",
    GridChar(NoLine, NoLine, DoubleLine, ThinLine): "╖",
    GridChar(ThinLine, NoLine, NoLine, DoubleLine): "╛",
    GridChar(DoubleLine, NoLine, NoLine, ThinLine): "╜",
    GridChar(ThinLine, DoubleLine, NoLine, NoLine): "╘",
    GridChar(NoLine, ThinLine, DoubleLine, NoLine): "╓",
    GridChar(NoLine, NoLine, ThinLine, DoubleLine): "╕",
    GridChar(ThinLine, DoubleLine, ThinLine, NoLine): "╞",
    GridChar(NoLine, ThinLine, DoubleLine, ThinLine): "╥",
    GridChar(ThinLine, NoLine, ThinLine, DoubleLine): "╡",
    GridChar(DoubleLine, ThinLine, NoLine, ThinLine): "╨",
    GridChar(DoubleLine, ThinLine, DoubleLine, NoLine): "╟",
    GridChar(NoLine, DoubleLine, ThinLine, DoubleLine): "╤",
    GridChar(DoubleLine, NoLine, DoubleLine, ThinLine): "╢",
    GridChar(ThinLine, DoubleLine, NoLine, DoubleLine): "╧",
    # ThickLine
    GridChar(ThickLine, ThickLine, ThickLine, ThickLine): "╋",
    GridChar(ThickLine, NoLine, NoLine, NoLine): "╹",
    GridChar(NoLine, ThickLine, NoLine, NoLine): "╺",
    GridChar(NoLine, NoLine, ThickLine, NoLine): "╻",
    GridChar(NoLine, NoLine, NoLine, ThickLine): "╸",
    GridChar(ThickLine, NoLine, ThickLine, NoLine): "┃",
    GridChar(NoLine, ThickLine, NoLine, ThickLine): "━",
    GridChar(ThickLine, ThickLine, NoLine, NoLine): "┗",
    GridChar(NoLine, ThickLine, ThickLine, NoLine): "┏",
    GridChar(NoLine, ThickLine, ThickLine, NoLine): "┏",
    GridChar(NoLine, NoLine, ThickLine, ThickLine): "┓",
    GridChar(ThickLine, NoLine, NoLine, ThickLine): "┛",
    GridChar(ThickLine, ThickLine, ThickLine, NoLine): "┣",
    GridChar(NoLine, ThickLine, ThickLine, ThickLine): "┳",
    GridChar(ThickLine, NoLine, ThickLine, ThickLine): "┫",
    GridChar(ThickLine, ThickLine, NoLine, ThickLine): "┻",
    # ThickLine Dashes
    GridChar(NoLine, ThickDoubleDashedLine, NoLine, ThickDoubleDashedLine): "╍",
    GridChar(ThickDoubleDashedLine, NoLine, ThickDoubleDashedLine, NoLine): "╏",
    GridChar(NoLine, ThickTripleDashedLine, NoLine, ThickTripleDashedLine): "┅",
    GridChar(ThickTripleDashedLine, NoLine, ThickTripleDashedLine, NoLine): "┇",
    GridChar(NoLine, ThickQuadrupleDashedLine, NoLine, ThickQuadrupleDashedLine): "┉",
    GridChar(ThickQuadrupleDashedLine, NoLine, ThickQuadrupleDashedLine, NoLine): "┋",
    # ThickLine / ThinLine
    GridChar(NoLine,                ThickLine,              NoLine,                 ThinLine        ): "╼",
    GridChar(ThinLine,              NoLine,                 ThickLine,              NoLine          ): "╽",
    GridChar(NoLine,                ThinLine,               NoLine,                 ThickLine       ): "╾",
    GridChar(ThickLine,             NoLine,                 ThinLine,               NoLine          ): "╿",
    GridChar(ThickLine,             ThinLine,               NoLine,                 NoLine          ): "┖",
    GridChar(NoLine,                ThickLine,              ThinLine,               NoLine          ): "┍",
    GridChar(NoLine,                NoLine,                 ThickLine,              ThinLine        ): "┒",
    GridChar(ThinLine,              NoLine,                 NoLine,                 ThickLine       ): "┙",
    GridChar(ThickLine,             NoLine,                 NoLine,                 ThinLine        ): "┚",
    GridChar(ThinLine,              ThickLine,              NoLine,                 NoLine          ): "┕",
    GridChar(NoLine,                ThinLine,               ThickLine,              NoLine          ): "┎",
    GridChar(NoLine,                NoLine,                 ThinLine,               ThickLine       ): "┑",
    GridChar(ThickLine,             ThinLine,               ThinLine,               NoLine          ): "┞",
    GridChar(NoLine,                ThickLine,              ThinLine,               ThinLine        ): "┮",
    GridChar(ThinLine,              NoLine,                 ThickLine,              ThinLine        ): "┧",
    GridChar(ThinLine,              ThinLine,               NoLine,                 ThickLine       ): "┵",
    GridChar(ThickLine,             NoLine,                 ThinLine,               ThinLine        ): "┦",
    GridChar(ThinLine,              ThickLine,              NoLine,                 ThinLine        ): "┶",
    GridChar(ThinLine,              ThinLine,               ThickLine,              NoLine          ): "┟",
    GridChar(NoLine,                ThinLine,               ThinLine,               ThickLine       ): "┭",
    GridChar(ThickLine,             ThinLine,               NoLine,                 ThinLine        ): "┸",
    GridChar(ThinLine,              ThickLine,              ThinLine,               NoLine          ): "┝",
    GridChar(NoLine,                ThinLine,               ThickLine,              ThinLine        ): "┰",
    GridChar(ThinLine,              NoLine,                 ThinLine,               ThickLine       ): "┥",
    GridChar(ThickLine,             ThickLine,              ThinLine,               NoLine          ): "┡",
    GridChar(NoLine,                ThickLine,              ThickLine,              ThinLine        ): "┲",
    GridChar(ThinLine,              NoLine,                 ThickLine,              ThickLine       ): "┪",
    GridChar(ThickLine,             ThinLine,               NoLine,                 ThickLine       ): "┹",
    GridChar(ThickLine,             ThickLine,              NoLine,                 ThinLine        ): "┺",
    GridChar(ThinLine,              ThickLine,              ThickLine,              NoLine          ): "┢",
    GridChar(NoLine,                ThinLine,               ThickLine,              ThickLine       ): "┱",
    GridChar(ThickLine,             NoLine,                 ThinLine,               ThickLine       ): "┩",
    GridChar(ThickLine,             ThinLine,               ThickLine,              NoLine          ): "┠",
    GridChar(NoLine,                ThickLine,              ThinLine,               ThickLine       ): "┯",
    GridChar(ThickLine,             NoLine,                 ThickLine,              ThinLine        ): "┨",
    GridChar(ThinLine,              ThickLine,              NoLine,                 ThickLine       ): "┷",
    GridChar(ThickLine,             ThinLine,               ThinLine,               ThinLine        ): "╀",
    GridChar(ThinLine,              ThickLine,              ThinLine,               ThinLine        ): "┾",
    GridChar(ThinLine,              ThinLine,               ThickLine,              ThinLine        ): "╁",
    GridChar(ThinLine,              ThinLine,               ThinLine,               ThickLine       ): "┽",
    GridChar(ThickLine,             ThickLine,              ThinLine,               ThinLine        ): "╄",
    GridChar(ThinLine,              ThickLine,              ThickLine,              ThinLine        ): "╆",
    GridChar(ThinLine,              ThinLine,               ThickLine,              ThickLine       ): "╅",
    GridChar(ThickLine,             ThinLine,               ThinLine,               ThickLine       ): "╃",
    GridChar(ThinLine,              ThickLine,              ThinLine,               ThickLine       ): "┿",
    GridChar(ThickLine,             ThinLine,               ThickLine,              ThinLine        ): "╂",
    GridChar(ThickLine,             ThickLine,              ThickLine,              ThinLine        ): "╊",
    GridChar(ThinLine,              ThickLine,              ThickLine,              ThickLine       ): "╈",
    GridChar(ThickLine,             ThinLine,               ThickLine,              ThickLine       ): "╉",
    GridChar(ThickLine,             ThickLine,              ThinLine,               ThickLine       ): "╇",
    # UpperRightEighthLine
    GridChar(UpperRightEighthLine,  NoLine,                 UpperRightEighthLine,   NoLine              ): "▕",
    GridChar(NoLine,                UpperRightEighthLine,   NoLine,                 UpperRightEighthLine): "▔",
    GridChar(UpperRightEighthLine,  UpperRightEighthLine,   NoLine,                 NoLine              ): " ",
    GridChar(NoLine,                UpperRightEighthLine,   UpperRightEighthLine,   NoLine              ): "▕",
    GridChar(NoLine,                NoLine,                 UpperRightEighthLine,   UpperRightEighthLine): "🭾", #
    GridChar(UpperRightEighthLine,  NoLine,                 NoLine,                 UpperRightEighthLine): "▔",
    GridChar(UpperRightEighthLine,  UpperRightEighthLine,   UpperRightEighthLine,   NoLine              ): "▕",
    GridChar(NoLine,                UpperRightEighthLine,   UpperRightEighthLine,   UpperRightEighthLine): "▔",
    GridChar(UpperRightEighthLine,  NoLine,                 UpperRightEighthLine,   UpperRightEighthLine): "▕",
    GridChar(UpperRightEighthLine,  UpperRightEighthLine,   NoLine,                 UpperRightEighthLine): "▔",
    GridChar(UpperRightEighthLine,  UpperRightEighthLine,   UpperRightEighthLine,   UpperRightEighthLine): "▕",
    # LowerLeftEighthLine
    GridChar(LowerLeftEighthLine,   NoLine,                 LowerLeftEighthLine,    NoLine              ): "▏",
    GridChar(NoLine,                LowerLeftEighthLine,    NoLine,                 LowerLeftEighthLine ): "▁",
    GridChar(LowerLeftEighthLine,   LowerLeftEighthLine,    NoLine,                 NoLine              ): "🭼", #
    GridChar(NoLine,                LowerLeftEighthLine,    LowerLeftEighthLine,    NoLine              ): "▁",
    GridChar(NoLine,                NoLine,                 LowerLeftEighthLine,    LowerLeftEighthLine ): " ",
    GridChar(LowerLeftEighthLine,   NoLine,                 NoLine,                 LowerLeftEighthLine ): "▏",
    GridChar(LowerLeftEighthLine,   LowerLeftEighthLine,    LowerLeftEighthLine,    NoLine              ): "▏",
    GridChar(NoLine,                LowerLeftEighthLine,    LowerLeftEighthLine,    LowerLeftEighthLine ): "▁",
    GridChar(LowerLeftEighthLine,   NoLine,                 LowerLeftEighthLine,    LowerLeftEighthLine ): "▏",
    GridChar(LowerLeftEighthLine,   LowerLeftEighthLine,    NoLine,                 LowerLeftEighthLine ): "▁",
    GridChar(LowerLeftEighthLine,   LowerLeftEighthLine,    LowerLeftEighthLine,    LowerLeftEighthLine ): "▏",
    # Eighth Combos
    GridChar(NoLine,                LowerLeftEighthLine,    UpperRightEighthLine,   NoLine              ): " ",
    GridChar(LowerLeftEighthLine,   NoLine,                 NoLine,                 UpperRightEighthLine): " ",
    GridChar(NoLine,                UpperRightEighthLine,   LowerLeftEighthLine,    NoLine              ): "🭽", #
    GridChar(UpperRightEighthLine,  NoLine,                 NoLine,                 LowerLeftEighthLine ): "🭿", #

    # Eighth/ThinLine combos
    GridChar(LowerLeftEighthLine,   ThinLine,               LowerLeftEighthLine,    NoLine              ): "▏",
    GridChar(NoLine,                LowerLeftEighthLine,    ThinLine,               LowerLeftEighthLine ): "▁",
    GridChar(LowerLeftEighthLine,   NoLine,                 LowerLeftEighthLine,    ThinLine            ): "▏",
    GridChar(ThinLine,              LowerLeftEighthLine,    NoLine,                 LowerLeftEighthLine ): "▁",
    GridChar(UpperRightEighthLine,  ThinLine,               UpperRightEighthLine,   NoLine              ): "▕",
    GridChar(NoLine,                UpperRightEighthLine,   ThinLine,               UpperRightEighthLine): "▔",
    GridChar(UpperRightEighthLine,  NoLine,                 UpperRightEighthLine,   ThinLine            ): "▕",
    GridChar(ThinLine,              UpperRightEighthLine,   NoLine,                 UpperRightEighthLine): "▔",
    GridChar(NoLine,                NoLine,                 UpperRightEighthLine,   LowerLeftEighthLine ): "▁",
    GridChar(LowerLeftEighthLine,   UpperRightEighthLine,   NoLine,                 NoLine              ): "▔",

    # UpperRightQuarterLine
    GridChar(UpperRightQuarterLine, NoLine, UpperRightQuarterLine, NoLine): "🮇",
    GridChar(NoLine, UpperRightQuarterLine, NoLine, UpperRightQuarterLine): "🮂",
    GridChar(NoLine, UpperRightQuarterLine, UpperRightQuarterLine, NoLine): "🮇",
    GridChar(UpperRightQuarterLine, NoLine, NoLine, UpperRightQuarterLine): "🮂",
    GridChar(UpperRightQuarterLine, UpperRightQuarterLine, UpperRightQuarterLine, NoLine): "🮇",
    GridChar(UpperRightQuarterLine, UpperRightQuarterLine, NoLine, UpperRightQuarterLine): "🮂",
    # Corners
    GridChar(NoLine, LowerLeftEighthLine, UpperRightQuarterLine, NoLine): " ",
    GridChar(NoLine, NoLine, LowerLeftQuarterLine, LowerLeftEighthLine): " ",
    GridChar(LowerLeftQuarterLine, NoLine, NoLine, UpperRightEighthLine): " ",
    GridChar(UpperRightQuarterLine, UpperRightEighthLine, NoLine, NoLine): " ",

    # LowerLeftQuarterLine
    GridChar(LowerLeftQuarterLine, NoLine, LowerLeftQuarterLine, NoLine): "▎",
    GridChar(NoLine, LowerLeftQuarterLine, NoLine, LowerLeftQuarterLine): "▂",
    GridChar(NoLine, LowerLeftQuarterLine, LowerLeftQuarterLine, NoLine): "▂",
    GridChar(LowerLeftQuarterLine, NoLine, NoLine, LowerLeftQuarterLine): "▎",
    GridChar(NoLine, LowerLeftQuarterLine, LowerLeftQuarterLine, LowerLeftQuarterLine): "▂",
    GridChar(LowerLeftQuarterLine, NoLine, LowerLeftQuarterLine, LowerLeftQuarterLine): "▎",
    # Half/ThinLine combos
    GridChar(LowerLeftQuarterLine, ThinLine, LowerLeftQuarterLine, NoLine): "▎",
    GridChar(NoLine, LowerLeftQuarterLine, ThinLine, LowerLeftQuarterLine): "▂",
    GridChar(LowerLeftQuarterLine, NoLine, LowerLeftQuarterLine, ThinLine): "▎",
    GridChar(ThinLine, LowerLeftQuarterLine, NoLine, LowerLeftQuarterLine): "▂",
    GridChar(UpperRightQuarterLine, ThinLine, UpperRightQuarterLine, NoLine): "🮇",
    GridChar(NoLine, UpperRightQuarterLine, ThinLine, UpperRightQuarterLine): "🮂",
    GridChar(UpperRightQuarterLine, NoLine, UpperRightQuarterLine, ThinLine): "🮇",
    GridChar(ThinLine, UpperRightQuarterLine, NoLine, UpperRightQuarterLine): "🮂",

    # UpperRightHalfLine
    GridChar(UpperRightHalfLine, NoLine, UpperRightHalfLine, NoLine): "▐",
    GridChar(NoLine, UpperRightHalfLine, NoLine, UpperRightHalfLine): "▀",
    GridChar(UpperRightHalfLine, UpperRightHalfLine, NoLine, NoLine): "▝",
    GridChar(NoLine, UpperRightHalfLine, UpperRightHalfLine, NoLine): "▐",
    GridChar(NoLine, NoLine, UpperRightHalfLine, UpperRightHalfLine): "▜",
    GridChar(UpperRightHalfLine, NoLine, NoLine, UpperRightHalfLine): "▀",
    GridChar(UpperRightHalfLine, UpperRightHalfLine, UpperRightHalfLine, NoLine): "▐",
    GridChar(NoLine, UpperRightHalfLine, UpperRightHalfLine, UpperRightHalfLine): "▜",
    GridChar(UpperRightHalfLine, NoLine, UpperRightHalfLine, UpperRightHalfLine): "▜",
    GridChar(UpperRightHalfLine, UpperRightHalfLine, NoLine, UpperRightHalfLine): "▀",
    GridChar(UpperRightHalfLine, UpperRightHalfLine, UpperRightHalfLine, UpperRightHalfLine): "▜",
    # LowerLeftHalfLine
    GridChar(LowerLeftHalfLine, NoLine, LowerLeftHalfLine, NoLine): "▌",
    GridChar(NoLine, LowerLeftHalfLine, NoLine, LowerLeftHalfLine): "▄",
    GridChar(LowerLeftHalfLine, LowerLeftHalfLine, NoLine, NoLine): "▙",
    GridChar(NoLine, LowerLeftHalfLine, LowerLeftHalfLine, NoLine): "▄",
    GridChar(NoLine, NoLine, LowerLeftHalfLine, LowerLeftHalfLine): "▖",
    GridChar(LowerLeftHalfLine, NoLine, NoLine, LowerLeftHalfLine): "▌",
    GridChar(LowerLeftHalfLine, LowerLeftHalfLine, LowerLeftHalfLine, NoLine): "▙",
    GridChar(NoLine, LowerLeftHalfLine, LowerLeftHalfLine, LowerLeftHalfLine): "▄",
    GridChar(LowerLeftHalfLine, NoLine, LowerLeftHalfLine, LowerLeftHalfLine): "▌",
    GridChar(LowerLeftHalfLine, LowerLeftHalfLine, NoLine, LowerLeftHalfLine): "▙",
    GridChar(LowerLeftHalfLine, LowerLeftHalfLine, LowerLeftHalfLine, LowerLeftHalfLine): "▙",
    # Half Combos
    GridChar(NoLine, UpperRightHalfLine, LowerLeftHalfLine, NoLine): "▛",
    GridChar(UpperRightHalfLine, NoLine, NoLine, LowerLeftHalfLine): "▟",
    GridChar(LowerLeftHalfLine, NoLine, NoLine, UpperRightHalfLine): "▘",
    GridChar(NoLine, LowerLeftHalfLine, UpperRightHalfLine, NoLine): "▗",
    # Half/ThinLine combos
    GridChar(LowerLeftHalfLine, ThinLine, LowerLeftHalfLine, NoLine): "▌",
    GridChar(NoLine, LowerLeftHalfLine, ThinLine, LowerLeftHalfLine): "▄",
    GridChar(LowerLeftHalfLine, NoLine, LowerLeftHalfLine, ThinLine): "▌",
    GridChar(ThinLine, LowerLeftHalfLine, NoLine, LowerLeftHalfLine): "▄",
    GridChar(UpperRightHalfLine, ThinLine, UpperRightHalfLine, NoLine): "▐",
    GridChar(NoLine, UpperRightHalfLine, ThinLine, UpperRightHalfLine): "▀",
    GridChar(UpperRightHalfLine, NoLine, UpperRightHalfLine, ThinLine): "▐",
    GridChar(ThinLine, UpperRightHalfLine, NoLine, UpperRightHalfLine): "▀",
    GridChar(UpperRightHalfLine, UpperRightHalfLine, LowerLeftHalfLine, LowerLeftHalfLine): "█",
    GridChar(LowerLeftHalfLine, UpperRightHalfLine, UpperRightHalfLine, LowerLeftHalfLine): "█",
    GridChar(LowerLeftHalfLine, LowerLeftHalfLine, UpperRightHalfLine, UpperRightHalfLine): "█",
    GridChar(UpperRightHalfLine, LowerLeftHalfLine, LowerLeftHalfLine, UpperRightHalfLine): "█",
    #
    GridChar(ThinLine, UpperRightHalfLine, LowerLeftHalfLine, ThinLine): "▛",
    GridChar(NoLine, UpperRightHalfLine, LowerLeftHalfLine, ThinLine): "▛",
    GridChar(ThinLine, UpperRightHalfLine, LowerLeftHalfLine, NoLine): "▛",
    GridChar(NoLine, UpperRightHalfLine, LowerLeftHalfLine, NoLine): "▛",
    GridChar(ThinLine, ThinLine, UpperRightHalfLine, UpperRightHalfLine): "▜",
    GridChar(NoLine, ThinLine, UpperRightHalfLine, UpperRightHalfLine): "▜",
    GridChar(ThinLine, NoLine, UpperRightHalfLine, UpperRightHalfLine): "▜",
    GridChar(NoLine, NoLine, UpperRightHalfLine, UpperRightHalfLine): "▜",
    GridChar(UpperRightHalfLine, ThinLine, ThinLine, LowerLeftHalfLine): "▟",
    GridChar(UpperRightHalfLine, NoLine, ThinLine, LowerLeftHalfLine): "▟",
    GridChar(UpperRightHalfLine, ThinLine, NoLine, LowerLeftHalfLine): "▟",
    GridChar(UpperRightHalfLine, NoLine, NoLine, LowerLeftHalfLine): "▟",
    GridChar(LowerLeftHalfLine, LowerLeftHalfLine, ThinLine, ThinLine): "▙",
    GridChar(LowerLeftHalfLine, LowerLeftHalfLine, NoLine, ThinLine): "▙",
    GridChar(LowerLeftHalfLine, LowerLeftHalfLine, ThinLine, NoLine): "▙",
    GridChar(LowerLeftHalfLine, LowerLeftHalfLine, NoLine, NoLine): "▙",
    # FullLine
    GridChar(FullLine, NoLine, FullLine, NoLine): "█",
    GridChar(NoLine, FullLine, NoLine, FullLine): "█",
    GridChar(FullLine, FullLine, NoLine, NoLine): "█",
    GridChar(NoLine, FullLine, FullLine, NoLine): "█",
    GridChar(NoLine, NoLine, FullLine, FullLine): "█",
    GridChar(FullLine, NoLine, NoLine, FullLine): "█",
    GridChar(FullLine, FullLine, FullLine, NoLine): "█",
    GridChar(NoLine, FullLine, FullLine, FullLine): "█",
    GridChar(FullLine, NoLine, FullLine, FullLine): "█",
    GridChar(FullLine, FullLine, NoLine, FullLine): "█",
    GridChar(FullLine, FullLine, FullLine, FullLine): "█",
    # FullLine End-stops
    GridChar(FullLine, NoLine, NoLine, NoLine): "▀",
    GridChar(NoLine, FullLine, NoLine, NoLine): "▐",
    GridChar(NoLine, NoLine, FullLine, NoLine): "▄",
    GridChar(NoLine, NoLine, NoLine, FullLine): "▌",
    # Full + Halves
    GridChar(FullLine, NoLine, NoLine, LowerLeftHalfLine): "█",
    GridChar(FullLine, LowerLeftHalfLine, NoLine, NoLine): "█",
    GridChar(NoLine, UpperRightHalfLine, FullLine, NoLine): "█",
    GridChar(NoLine, NoLine, FullLine, UpperRightHalfLine): "█",
    GridChar(FullLine, NoLine, NoLine, UpperRightHalfLine): "▀",
    GridChar(FullLine, UpperRightHalfLine, NoLine, NoLine): "▀",
    GridChar(NoLine, LowerLeftHalfLine, FullLine, NoLine): "▄",
    GridChar(NoLine, NoLine, FullLine, LowerLeftHalfLine): "▄",
    GridChar(NoLine, FullLine, LowerLeftHalfLine, NoLine): "█",
    GridChar(NoLine, NoLine, LowerLeftHalfLine, FullLine): "▌",
    GridChar(LowerLeftHalfLine, NoLine, NoLine, FullLine): "▌",
    GridChar(LowerLeftHalfLine, FullLine, NoLine, NoLine): "█",
    GridChar(FullLine, LowerLeftHalfLine, FullLine, NoLine): "█",
    GridChar(NoLine, LowerLeftHalfLine, FullLine, LowerLeftHalfLine): "▄",
    GridChar(FullLine, LowerLeftHalfLine, FullLine, LowerLeftHalfLine): "█",
    GridChar(FullLine, UpperRightHalfLine, FullLine, UpperRightHalfLine): "█",
    GridChar(FullLine, NoLine, FullLine, LowerLeftHalfLine): "█",
    GridChar(FullLine, UpperRightHalfLine, NoLine, UpperRightHalfLine): "▀",
    GridChar(FullLine, UpperRightHalfLine, FullLine, NoLine): "█",
    GridChar(FullLine, UpperRightHalfLine, FullLine, UpperRightHalfLine): "█",
    GridChar(NoLine, UpperRightHalfLine, FullLine, UpperRightHalfLine): "█",
    GridChar(UpperRightHalfLine, NoLine, UpperRightHalfLine, FullLine): "█",
    GridChar(FullLine, LowerLeftHalfLine, NoLine, LowerLeftHalfLine): "█",
    GridChar(LowerLeftHalfLine, FullLine, LowerLeftHalfLine, NoLine): "█",
    GridChar(NoLine, NoLine, UpperRightHalfLine, FullLine): "█",
    GridChar(UpperRightHalfLine, NoLine, NoLine, FullLine): "█",
    GridChar(NoLine, FullLine, LowerLeftHalfLine, FullLine): "█",
    GridChar(NoLine, FullLine, UpperRightHalfLine, FullLine): "█",
    GridChar(UpperRightHalfLine, FullLine, UpperRightHalfLine, FullLine): "█",
    GridChar(UpperRightHalfLine, FullLine, LowerLeftHalfLine, FullLine): "█",
    GridChar(LowerLeftHalfLine, FullLine, LowerLeftHalfLine, FullLine): "█",
    GridChar(LowerLeftHalfLine, FullLine, NoLine, FullLine): "█",

    # Full + Eighths
    GridChar(NoLine, UpperRightEighthLine, FullLine, UpperRightEighthLine): "█",
    GridChar(UpperRightEighthLine, NoLine, UpperRightEighthLine, FullLine): "█",
    GridChar(FullLine, LowerLeftEighthLine, NoLine, LowerLeftEighthLine): "█",
    GridChar(LowerLeftEighthLine, FullLine, LowerLeftEighthLine, NoLine): "█",

    # UpperRightHalfDottedLine
    GridChar(UpperRightHalfDottedLine, NoLine, UpperRightHalfDottedLine, NoLine): "⢸",
    GridChar(NoLine, UpperRightHalfDottedLine, NoLine, UpperRightHalfDottedLine): "⠛",
    GridChar(UpperRightHalfDottedLine, UpperRightHalfDottedLine, NoLine, NoLine): "▝",
    GridChar(NoLine, UpperRightHalfDottedLine, UpperRightHalfDottedLine, NoLine): "⢸",
    GridChar(NoLine, NoLine, UpperRightHalfDottedLine, UpperRightHalfDottedLine): "▜",
    GridChar(UpperRightHalfDottedLine, NoLine, NoLine, UpperRightHalfDottedLine): "⠛",
    GridChar(UpperRightHalfDottedLine, UpperRightHalfDottedLine, UpperRightHalfDottedLine, NoLine): "⢸",
    GridChar(NoLine, UpperRightHalfDottedLine, UpperRightHalfDottedLine, UpperRightHalfDottedLine): "▜",
    GridChar(UpperRightHalfDottedLine, NoLine, UpperRightHalfDottedLine, UpperRightHalfDottedLine): "▜",
    GridChar(UpperRightHalfDottedLine, UpperRightHalfDottedLine, NoLine, UpperRightHalfDottedLine): "⠛",
    GridChar(UpperRightHalfDottedLine, UpperRightHalfDottedLine, UpperRightHalfDottedLine, UpperRightHalfDottedLine): "▜",
    # LowerLeftHalfDottedLine
    GridChar(LowerLeftHalfDottedLine, NoLine, LowerLeftHalfDottedLine, NoLine): "⡇",
    GridChar(NoLine, LowerLeftHalfDottedLine, NoLine, LowerLeftHalfDottedLine): "⣤",
    GridChar(LowerLeftHalfDottedLine, LowerLeftHalfDottedLine, NoLine, NoLine): "▙",
    GridChar(NoLine, LowerLeftHalfDottedLine, LowerLeftHalfDottedLine, NoLine): "⣤",
    GridChar(NoLine, NoLine, LowerLeftHalfDottedLine, LowerLeftHalfDottedLine): "▖",
    GridChar(LowerLeftHalfDottedLine, NoLine, NoLine, LowerLeftHalfDottedLine): "⡇",
    GridChar(LowerLeftHalfDottedLine, LowerLeftHalfDottedLine, LowerLeftHalfDottedLine, NoLine): "▙",
    GridChar(NoLine, LowerLeftHalfDottedLine, LowerLeftHalfDottedLine, LowerLeftHalfDottedLine): "⣤",
    GridChar(LowerLeftHalfDottedLine, NoLine, LowerLeftHalfDottedLine, LowerLeftHalfDottedLine): "⡇",
    GridChar(LowerLeftHalfDottedLine, LowerLeftHalfDottedLine, NoLine, LowerLeftHalfDottedLine): "▙",
    GridChar(LowerLeftHalfDottedLine, LowerLeftHalfDottedLine, LowerLeftHalfDottedLine, LowerLeftHalfDottedLine): "▙",
    # Half Combos
    GridChar(NoLine, UpperRightHalfDottedLine, LowerLeftHalfDottedLine, NoLine): "▛",
    GridChar(UpperRightHalfDottedLine, NoLine, NoLine, LowerLeftHalfDottedLine): "▟",
    GridChar(LowerLeftHalfDottedLine, NoLine, NoLine, UpperRightHalfDottedLine): "▘",
    GridChar(NoLine, LowerLeftHalfDottedLine, UpperRightHalfDottedLine, NoLine): "▗",
    # FullDottedLine
    GridChar(FullDottedLine, NoLine, FullDottedLine, NoLine): "⣿",
    GridChar(NoLine, FullDottedLine, NoLine, FullDottedLine): "⣿",
    GridChar(FullDottedLine, FullDottedLine, NoLine, NoLine): "⣿",
    GridChar(NoLine, FullDottedLine, FullDottedLine, NoLine): "⣿",
    GridChar(NoLine, NoLine, FullDottedLine, FullDottedLine): "⣿",
    GridChar(FullDottedLine, NoLine, NoLine, FullDottedLine): "⣿",
    GridChar(FullDottedLine, FullDottedLine, FullDottedLine, NoLine): "⣿",
    GridChar(NoLine, FullDottedLine, FullDottedLine, FullDottedLine): "⣿",
    GridChar(FullDottedLine, NoLine, FullDottedLine, FullDottedLine): "⣿",
    GridChar(FullDottedLine, FullDottedLine, NoLine, FullDottedLine): "⣿",
    GridChar(FullDottedLine, FullDottedLine, FullDottedLine, FullDottedLine): "⣿",
    # FullDottedLine End-stops
    GridChar(FullDottedLine, NoLine, NoLine, NoLine): "⠛",
    GridChar(NoLine, FullDottedLine, NoLine, NoLine): "⢸",
    GridChar(NoLine, NoLine, FullDottedLine, NoLine): "⣤",
    GridChar(NoLine, NoLine, NoLine, FullDottedLine): "⡇",
    # Full + Halves
    GridChar(FullDottedLine, NoLine, NoLine, LowerLeftHalfDottedLine): "⣿",
    GridChar(FullDottedLine, LowerLeftHalfDottedLine, NoLine, NoLine): "⣿",
    GridChar(NoLine, UpperRightHalfDottedLine, FullDottedLine, NoLine): "⣿",
    GridChar(NoLine, NoLine, FullDottedLine, UpperRightHalfDottedLine): "⣿",
    GridChar(FullDottedLine, NoLine, NoLine, UpperRightHalfDottedLine): "⠛",
    GridChar(FullDottedLine, UpperRightHalfDottedLine, NoLine, NoLine): "⠛",
    GridChar(NoLine, LowerLeftHalfDottedLine, FullDottedLine, NoLine): "⣤",
    GridChar(NoLine, NoLine, FullDottedLine, LowerLeftHalfDottedLine): "⣤",
    GridChar(NoLine, FullDottedLine, LowerLeftHalfDottedLine, NoLine): "⣿",
    GridChar(NoLine, NoLine, LowerLeftHalfDottedLine, FullDottedLine): "⡇",
    GridChar(LowerLeftHalfDottedLine, NoLine, NoLine, FullDottedLine): "⡇",
    GridChar(LowerLeftHalfDottedLine, FullDottedLine, NoLine, NoLine): "⣿",
    GridChar(FullDottedLine, LowerLeftHalfDottedLine, FullDottedLine, NoLine): "⣿",
    GridChar(NoLine, LowerLeftHalfDottedLine, FullDottedLine, LowerLeftHalfDottedLine): "⣤",
    GridChar(FullDottedLine, LowerLeftHalfDottedLine, FullDottedLine, LowerLeftHalfDottedLine): "⣿",
    GridChar(FullDottedLine, UpperRightHalfDottedLine, FullDottedLine, UpperRightHalfDottedLine): "⣿",
    GridChar(FullDottedLine, NoLine, FullDottedLine, LowerLeftHalfDottedLine): "⣿",
    GridChar(FullDottedLine, UpperRightHalfDottedLine, NoLine, UpperRightHalfDottedLine): "⠛",
    GridChar(FullDottedLine, UpperRightHalfDottedLine, FullDottedLine, NoLine): "⣿",
    GridChar(FullDottedLine, UpperRightHalfDottedLine, FullDottedLine, UpperRightHalfDottedLine): "⣿",
    GridChar(NoLine, UpperRightHalfDottedLine, FullDottedLine, UpperRightHalfDottedLine): "⣿",
    GridChar(UpperRightHalfDottedLine, NoLine, UpperRightHalfDottedLine, FullDottedLine): "⣿",
    GridChar(FullDottedLine, LowerLeftHalfDottedLine, NoLine, LowerLeftHalfDottedLine): "⣿",
    GridChar(LowerLeftHalfDottedLine, FullDottedLine, LowerLeftHalfDottedLine, NoLine): "⣿",
    GridChar(NoLine, NoLine, UpperRightHalfDottedLine, FullDottedLine): "⣿",
    GridChar(UpperRightHalfDottedLine, NoLine, NoLine, FullDottedLine): "⣿",
    GridChar(NoLine, FullDottedLine, LowerLeftHalfDottedLine, FullDottedLine): "⣿",
    GridChar(NoLine, FullDottedLine, UpperRightHalfDottedLine, FullDottedLine): "⣿",
    GridChar(UpperRightHalfDottedLine, FullDottedLine, UpperRightHalfDottedLine, FullDottedLine): "⣿",
    GridChar(UpperRightHalfDottedLine, FullDottedLine, LowerLeftHalfDottedLine, FullDottedLine): "⣿",
    GridChar(LowerLeftHalfDottedLine, FullDottedLine, LowerLeftHalfDottedLine, FullDottedLine): "⣿",
    GridChar(LowerLeftHalfDottedLine, FullDottedLine, NoLine, FullDottedLine): "⣿",

}
# fmt: off

@lru_cache
def get_grid_char(key: GridChar) -> str:
    """Return the character represented by a combination of :class:`LineStyles`."""
    if key in _GRID_CHARS:
        return _GRID_CHARS[key]
    else:
        # If there is no matching character representation, replace the line style
        # whose parent has the highest ranking with the parent style until a character
        # is found
        m_key = list(key)
        while any(x.parent for x in m_key):
            parent_ranks = {(9999,) if line.parent == NoLine else line.parent.rank: i for i, line in enumerate(m_key) if line.parent}
            idx = parent_ranks[max(parent_ranks)]
            if parent := m_key[idx].parent:
                m_key[idx] = parent
                char = _GRID_CHARS.get(GridChar(*m_key))
                if char:
                    return char
        # If all else fails, return a space
        else:
            return " "


def _combine_grids(grid_a: GridStyle, grid_b: GridStyle) -> GridStyle:
    """Combine a pair of grids."""
    grid_style = GridStyle()
    grid_style.grid = {}
    for part in GridPart:
        grid_style.grid[part] = GridChar(
            *(max(grid_a.grid[part][i], grid_b.grid[part][i]) for i in range(4))
        )
    return grid_style


class GridStyle:
    """A collection of characters which can be used to draw a grid."""

    class _BorderLineChars(NamedTuple):
        LEFT: str
        MID: str
        SPLIT: str
        RIGHT: str

    _grid_cache: FastDictCache[tuple[GridStyle, GridStyle], GridStyle] = FastDictCache(get_value=_combine_grids)

    def __init__(
        self, line_style: LineStyle = NoLine, mask: Mask = Masks.grid
    ) -> None:
        """Create a new :py:class:`GridStyle` instance.

        Args:
            line_style: The line style to use to construct the grid
            mask: A mask which can be used to exclude certain character from the grid
        """
        self.grid = {
            part: GridChar(
                *((line_style if x else NoLine) for x in mask.mask[part])
            )
            for part in GridPart
        }
        self.mask = mask

    @property
    def TOP(self) -> _BorderLineChars:
        """Allow dotted attribute access to the top grid row."""
        return self._BorderLineChars(
            get_grid_char(self.grid[GridPart.TOP_LEFT]),
            get_grid_char(self.grid[GridPart.TOP_MID]),
            get_grid_char(self.grid[GridPart.TOP_SPLIT]),
            get_grid_char(self.grid[GridPart.TOP_RIGHT]),
        )

    @property
    def MID(self) -> _BorderLineChars:
        """Allow dotted attribute access to the mid grid row."""
        return self._BorderLineChars(
            get_grid_char(self.grid[GridPart.MID_LEFT]),
            get_grid_char(self.grid[GridPart.MID_MID]),
            get_grid_char(self.grid[GridPart.MID_SPLIT]),
            get_grid_char(self.grid[GridPart.MID_RIGHT]),
        )

    @property
    def SPLIT(self) -> _BorderLineChars:
        """Allow dotted attribute access to the split grid row."""
        return self._BorderLineChars(
            get_grid_char(self.grid[GridPart.SPLIT_LEFT]),
            get_grid_char(self.grid[GridPart.SPLIT_MID]),
            get_grid_char(self.grid[GridPart.SPLIT_SPLIT]),
            get_grid_char(self.grid[GridPart.SPLIT_RIGHT]),
        )

    @property
    def BOTTOM(self) -> _BorderLineChars:
        """Allow dotted attribute access to the bottom grid row."""
        return self._BorderLineChars(
            get_grid_char(self.grid[GridPart.BOTTOM_LEFT]),
            get_grid_char(self.grid[GridPart.BOTTOM_MID]),
            get_grid_char(self.grid[GridPart.BOTTOM_SPLIT]),
            get_grid_char(self.grid[GridPart.BOTTOM_RIGHT]),
        )

    @property
    def HORIZONTAL(self) -> str:
        """For compatibility with :class:`prompt_toolkit.widgets.base.Border`."""
        return self.SPLIT_MID

    @property
    def VERTICAL(self) -> str:
        """For compatibility with :class:`prompt_toolkit.widgets.base.Border`."""
        return self.MID_SPLIT

    def __getattr__(self, value: str) -> str:
        """Allow the parts of the grid to be accessed as attributes by name."""
        key = getattr(GridPart, value)
        return get_grid_char(self.grid[key])

    def __dir__(self) -> list:
        """Lit the public attributes of the grid style."""
        return [x.name for x in GridPart] + ["TOP", "MID", "SPLIT", "BOTTOM"]

    def __add__(self, other: GridStyle) -> GridStyle:
        """Combine two grid styles."""
        return self._grid_cache[self, other]

    def __repr__(self) -> str:
        """Return a string representation of the grid style."""
        return "\n".join(
            "".join(
                get_grid_char(char_key)
                for char_key in list(self.grid.values())[i * 4 : (i + 1) * 4]
            )
            for i in range(4)
        )

ThinGrid = ThinLine.grid

InsetGrid = (
    UpperRightQuarterLine.left_edge
    + LowerLeftQuarterLine.right_edge
    + UpperRightEighthLine.bottom_edge
    + LowerLeftEighthLine.top_edge
    + ThinLine.inner
)

OutsetGrid = (
    LowerLeftEighthLine.top_edge
    + UpperRightEighthLine.right_edge
    + UpperRightEighthLine.bottom_edge
    + LowerLeftEighthLine.left_edge
    + ThinLine.inner
)

OuterHalfGrid = (
    UpperRightHalfLine.top_edge
    + UpperRightHalfLine.right_edge
    + LowerLeftHalfLine.left_edge
    + LowerLeftHalfLine.bottom_edge
    + ThinLine.inner
)

InnerHalfGrid = (
    LowerLeftHalfLine.top_edge
    + LowerLeftHalfLine.right_edge
    + UpperRightHalfLine.left_edge
    + UpperRightHalfLine.bottom_edge
    + ThinLine.inner
)


class DiLineStyle(NamedTuple):
    """A description of a cell border: a :class:`LineStyle` for each edge."""

    top: LineStyle = NoLine
    right: LineStyle = NoLine
    bottom: LineStyle = NoLine
    left: LineStyle = NoLine

    @classmethod
    def from_value(cls, value: LineStyle) -> DiLineStyle:
        """Construct an instance from a single value."""
        return cls(top=value, right=value, bottom=value, left=value)



"""Responsible for loading data from urls."""

from __future__ import annotations

import base64
import binascii
import io
import logging
from pathlib import Path
from typing import TYPE_CHECKING, overload
from urllib.parse import unquote, urljoin, urlunsplit

import upath
from aiohttp.client_reqrep import ClientResponse
from fsspec.implementations.cached import WholeFileCacheFileSystem
from fsspec.implementations.http import HTTPFileSystem as FsHTTPFileSystem
from fsspec.registry import register_implementation as fs_register_implementation
from upath import UPath
from upath.implementations.http import HTTPPath as _HTTPPath
from upath.implementations.http import _HTTPAccessor
from upath.registry import _registry

if TYPE_CHECKING:
    from io import BufferedRandom, BufferedReader, BufferedWriter, FileIO, TextIOWrapper
    from os import PathLike
    from typing import IO, Any, BinaryIO, Literal
    from urllib.parse import SplitResult

    from _typeshed import (
        OpenBinaryMode,
        OpenBinaryModeReading,
        OpenBinaryModeUpdating,
        OpenBinaryModeWriting,
        OpenTextMode,
    )
    from upath.core import PT


log = logging.getLogger(__name__)


def parse_path(path: str | PathLike, resolve: bool = True) -> Path:
    """Parse and resolve a path."""
    if not isinstance(path, Path):
        path = UPath(path)
    try:
        path = path.expanduser()
    except NotImplementedError:
        pass
    if resolve:
        try:
            path = path.resolve()
        except (AttributeError, NotImplementedError, Exception):
            log.info("Path %s not resolvable", path)
            pass
    return path


def _raise_for_status(self: ClientResponse) -> None:
    """Monkey-patch :py:class:`aiohttp.ClientResponse` not to raise for any status."""
    pass


setattr(ClientResponse, "raise_for_status", _raise_for_status)  # noqa B010


class DataPath(upath.core.UPath):
    """A :py:class:`pathlib` compatible class for reading data URIs."""

    @overload
    def open(
        self,
        mode: OpenTextMode = "r",
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> TextIOWrapper:
        ...

    # Unbuffered binary mode: returns a FileIO
    @overload
    def open(
        self,
        mode: OpenBinaryMode,
        buffering: Literal[0],
        encoding: None = None,
        errors: None = None,
        newline: None = None,
    ) -> FileIO:
        ...

    # Buffering is on: return BufferedRandom, BufferedReader, or BufferedWriter
    @overload
    def open(
        self,
        mode: OpenBinaryModeUpdating,
        buffering: Literal[-1, 1] = -1,
        encoding: None = None,
        errors: None = None,
        newline: None = None,
    ) -> BufferedRandom:
        ...

    @overload
    def open(
        self,
        mode: OpenBinaryModeWriting,
        buffering: Literal[-1, 1] = -1,
        encoding: None = None,
        errors: None = None,
        newline: None = None,
    ) -> BufferedWriter:
        ...

    @overload
    def open(
        self,
        mode: OpenBinaryModeReading,
        buffering: Literal[-1, 1] = -1,
        encoding: None = None,
        errors: None = None,
        newline: None = None,
    ) -> BufferedReader:
        ...

    # Buffering cannot be determined: fall back to BinaryIO
    @overload
    def open(
        self,
        mode: OpenBinaryMode,
        buffering: int = -1,
        encoding: None = None,
        errors: None = None,
        newline: None = None,
    ) -> BinaryIO:
        ...

    # Fallback if mode is not specified
    @overload
    def open(
        self,
        mode: str,
        buffering: int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> IO[Any]:
        ...

    def open(
        self,
        mode: OpenTextMode
        | OpenBinaryMode
        | OpenBinaryModeReading
        | OpenBinaryModeWriting
        | OpenBinaryModeUpdating
        | str = "r",
        buffering: Literal[-1, 0, 1] | int = -1,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> IO[Any]:
        """Return an io object for the data in the URI."""
        assert self._url is not None
        data_format, _, encoded_data = self._url.path.partition(",")
        _mime, *params = data_format.split(";")

        data_bytes = None
        data_str = None

        if "base64" in params:
            try:
                data_bytes = base64.b64decode(encoded_data)
            except binascii.Error:
                log.warning("Failed to decode base64 encoded data")
                data_bytes = b""
        else:
            data_str = unquote(encoded_data)

        if "b" in mode:
            if data_bytes is None:
                assert data_str is not None
                data_bytes = data_str.encode()
            return io.BytesIO(data_bytes)
        else:
            if data_str is None:
                assert data_bytes is not None
                decode_kwargs: dict[str, str] = {}
                if encoding is not None:  #:= kwargs.get("encoding"):
                    decode_kwargs["encoding"] = str(encoding)
                data_str = data_bytes.decode(**decode_kwargs)
            return io.StringIO(data_str)

    def exists(self, **kwargs: Any) -> bool:
        """Affirm tat data URIs always exist."""
        return True

    @property
    def _mime(self) -> str:
        """Return the media type of the data URI."""
        assert self._url is not None
        data_format, _, _encoded_data = self._url.path.partition(",")
        mime, *params = data_format.split(";")
        return mime


_registry.known_implementations["data"] = "euporie.core.path.DataPath"


class CachingHTTPAccessor(_HTTPAccessor):
    """A :py:module:`universal_pathlib` accessor which caches content."""

    def __init__(self, parsed_url: SplitResult | None, **kwargs: Any) -> None:
        """Load a caching filesystem."""
        cls = WholeFileCacheFileSystem
        assert parsed_url is not None
        url_kwargs = cls._get_kwargs_from_urls(urlunsplit(parsed_url))
        url_kwargs.update(kwargs)
        url_kwargs.setdefault("target_protocol", "http")
        # url_kwargs.setdefault(
        #     "cache_storage", str(Path(user_cache_dir("euporie")) / "web")
        # )
        self._fs = cls(**url_kwargs)

    def _format_path(self, path: upath.UPath) -> str:
        if (url := path._url) is not None:
            return url.geturl()
        return super()._format_path(path)


class HTTPPath(_HTTPPath):
    """An HTTP path which caches content."""

    _hash: int
    _default_accessor = CachingHTTPAccessor

    def __truediv__(self: PT, key: str | PathLike) -> PT:
        """Join a path to a HTTP URI."""
        return self.__class__(urljoin(str(self), str(key)))

    def __str__(self) -> str:
        """Represent the path as a string."""
        try:
            return self._str
        except AttributeError:
            if url := self._url:
                self._str = url.geturl()
            else:
                return super().__str__()
            return self._str

    def __hash__(self) -> int:
        """Provide a unique hash of the path."""
        try:
            return self._hash
        except AttributeError:
            self._hash = hash(self._url)
            return self._hash


_registry.known_implementations["http"] = "euporie.core.path.HTTPPath"
_registry.known_implementations["https"] = "euporie.core.path.HTTPPath"


class HTTPFileSystem(FsHTTPFileSystem):
    """A :py:class:`HTTPFileSystem` which does not raise exceptions on 404 errors."""

    def _raise_not_found_for_status(self, response: ClientResponse, url: str) -> None:
        """Do not raise an exception for 404 errors."""
        pass


fs_register_implementation("http", HTTPFileSystem, clobber=True)
fs_register_implementation("https", HTTPFileSystem, clobber=True)



"""A text base user interface for euporie."""

from __future__ import annotations

import asyncio
import logging
from functools import partial
from typing import TYPE_CHECKING, cast

from prompt_toolkit.filters import Condition
from prompt_toolkit.formatted_text.base import to_formatted_text
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    DynamicContainer,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import Dimension
from upath import UPath

from euporie.core import __logo__
from euporie.core.app import BaseApp
from euporie.core.commands import add_cmd, get_cmd
from euporie.core.config import add_setting
from euporie.core.ft.utils import truncate
from euporie.core.key_binding.registry import register_bindings
from euporie.core.tabs.base import Tab
from euporie.core.widgets.decor import Pattern
from euporie.core.widgets.dialog import (
    AboutDialog,
    ConfirmDialog,
    ErrorDialog,
    MsgBoxDialog,
    NoKernelsDialog,
    OpenFileDialog,
    SaveAsDialog,
    SelectKernelDialog,
    ShortcutsDialog,
    UnsavedDialog,
)
from euporie.core.widgets.file_browser import FileBrowser
from euporie.core.widgets.layout import TabBarControl, TabBarTab
from euporie.core.widgets.menu import MenuBar, MenuItem
from euporie.core.widgets.pager import Pager
from euporie.core.widgets.palette import CommandPalette
from euporie.core.widgets.search_bar import SearchBar
from euporie.core.widgets.status_bar import StatusBar
from euporie.notebook.enums import TabMode
from euporie.notebook.tabs import Notebook
from euporie.notebook.widgets.side_bar import SideBar

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from typing import Any, Callable, Sequence

    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.layout.containers import AnyContainer, Float

    from euporie.core.widgets.cell import Cell
    from euporie.core.widgets.status_bar import StatusBarFields

log = logging.getLogger(__name__)


class NotebookApp(BaseApp):
    """Notebook app.

    Interactively edit a notebook file.

    Launch the interactive TUI notebook editor, allowing you to run and edit Jupyter
    notebooks in the terminal.
    """

    name = "notebook"
    need_mouse_support = True

    def __init__(self, **kwargs: Any) -> None:
        """Create a new euporie text user interface application instance."""
        kwargs.setdefault("title", "euporie-notebook")
        kwargs.setdefault("full_screen", True)
        kwargs.setdefault("leave_graphics", False)
        super().__init__(**kwargs)
        self.bindings_to_load.append("euporie.notebook.app.NotebookApp")

        # Register config hooks
        self.config.get_item("show_cell_borders").event += lambda x: self.refresh()

    def statusbar_defaults(self) -> StatusBarFields | None:
        """Load the default statusbar fields (run after keybindings are loaded)."""
        return (
            [
                [
                    ("", "Press "),
                    ("bold", get_cmd("new-notebook").key_str()),
                    ("", " to start a new notebook"),
                ],
            ],
            [[("", "Press "), ("bold", get_cmd("quit").key_str()), ("", " to quit")]],
        )

    async def _poll_terminal_colors(self) -> None:
        """Repeatedly query the terminal for its background and foreground colours."""
        while self.config.terminal_polling_interval:
            await asyncio.sleep(self.config.terminal_polling_interval)
            self.term_info.colors.send()

    def post_load(self) -> None:
        """Continue loading the app."""
        super().post_load()
        # Ensure an opened tab is focused
        if self.tab:
            self.tab.focus()

        # Load style hooks and start polling terminal style
        if self.config.terminal_polling_interval and hasattr(
            self.input, "vt100_parser"
        ):
            self.create_background_task(self._poll_terminal_colors())

    def format_title(self) -> StyleAndTextTuples:
        """Format the tab's title for display in the top right of the app."""
        if self.tab:
            title = truncate(to_formatted_text(self.tab.title, style="bold"), 30)
            return [("", " "), *title, ("", " ")]
        else:
            return []

    def tab_container(self) -> AnyContainer:
        """Return a container with all opened tabs.

        Returns:
            A layout displaying the opened tab containers.

        """
        if self.tabs:
            if TabMode(self.config.tab_mode) == TabMode.TILE_HORIZONTALLY:
                return HSplit(
                    children=self.tabs,
                    padding=1,
                    padding_style="class:tab-padding",
                    padding_char="─",
                )
            elif TabMode(self.config.tab_mode) == TabMode.TILE_VERTICALLY:
                return VSplit(
                    children=self.tabs,
                    padding=1,
                    padding_style="class:tab-padding",
                    padding_char="│",
                )
            else:
                return DynamicContainer(lambda: self.tabs[self._tab_idx])
        else:
            return Pattern(
                self.config.background_character, self.config.background_pattern
            )

    def load_container(self) -> FloatContainer:
        """Build the main application layout."""
        have_tabs = Condition(lambda: bool(self.tabs))

        self.logo = Window(
            FormattedTextControl(
                [("", f" {__logo__} ")],
                focusable=~have_tabs,
                show_cursor=False,
                style="class:menu,logo",
            ),
            height=1,
            width=3,
            dont_extend_width=True,
        )

        title_bar = ConditionalContainer(
            Window(
                content=FormattedTextControl(self.format_title, show_cursor=False),
                height=1,
                style="class:menu.item",
                dont_extend_width=True,
                align=WindowAlign.RIGHT,
            ),
            filter=have_tabs,
        )

        self.tab_bar_control = TabBarControl(
            tabs=self.tab_bar_tabs,
            active=lambda: self._tab_idx,
            closeable=True,
        )
        tab_bar = ConditionalContainer(
            Window(
                self.tab_bar_control,
                height=2,
                style="class:app-tab-bar",
            ),
            filter=Condition(
                lambda: (len(self.tabs) > 1 or self.config.always_show_tab_bar)
                and TabMode(self.config.tab_mode) == TabMode.STACK
            ),
        )

        self.pager = Pager()
        self.search_bar = SearchBar()

        self.dialogs["command-palette"] = CommandPalette(self)
        self.dialogs["about"] = AboutDialog(self)
        self.dialogs["open-file"] = OpenFileDialog(self)
        self.dialogs["save-as"] = SaveAsDialog(self)
        self.dialogs["no-kernels"] = NoKernelsDialog(self)
        self.dialogs["change-kernel"] = SelectKernelDialog(self)
        self.dialogs["confirm"] = ConfirmDialog(self)
        self.dialogs["error"] = ErrorDialog(self)
        self.dialogs["unsaved"] = UnsavedDialog(self)
        self.dialogs["shortcuts"] = ShortcutsDialog(self)
        self.dialogs["msgbox"] = MsgBoxDialog(self)

        top_bar = ConditionalContainer(
            content=VSplit(
                [
                    self.logo,
                    MenuBar(app=self, menu_items=self.load_menu_items()),
                    title_bar,
                ]
            ),
            filter=self.config.filter("show_top_bar"),
        )

        titles: Sequence[str]
        icons: Sequence[str]
        panels: Sequence[AnyContainer]
        titles, icons, panels = list(
            zip(
                *[
                    (
                        "File Browser",
                        "",
                        FileBrowser(width=25, on_open=self.open_file),
                    )
                ]
            ),
        )
        self.side_bar = SideBar(titles, icons, panels)

        self.container = FloatContainer(
            content=HSplit(
                [
                    top_bar,
                    VSplit(
                        [
                            self.side_bar,
                            HSplit(
                                [
                                    tab_bar,
                                    DynamicContainer(self.tab_container),
                                    self.pager,
                                ],
                                width=Dimension(weight=1),
                            ),
                        ],
                        height=Dimension(min=1),
                    ),
                    self.search_bar,
                    StatusBar(default=self.statusbar_defaults()),
                ],
                style="class:body",
            ),
            floats=cast("list[Float]", self.floats),
        )

        return self.container

    def tab_bar_tabs(self) -> list[TabBarTab]:
        """Return a list of the current tabs for the tab-bar."""
        return [
            TabBarTab(
                title=partial(lambda x: x.title, tab),
                on_activate=partial(setattr, self, "tab_idx", i),
                on_close=partial(self.close_tab, tab),
            )
            for i, tab in enumerate(self.tabs)
        ]

    def _handle_exception(
        self, loop: AbstractEventLoop, context: dict[str, Any]
    ) -> None:
        exception = context.get("exception")
        # Also display a dialog to the user
        self.dialogs["error"].show(exception=exception)
        # Log observed exceptions to the log
        log.exception("An unhandled exception occurred", exc_info=exception)

    def exit(self, *args: Any, **kwargs: Any) -> None:
        """Check for unsaved files before closing.

        Creates a chain of close file commands, where the callback for each triggers
        the closure of the next. The closing process can be cancelled anywhere along
        the chain.

        Args:
            args: Positional arguments
            kwargs: Key word arguments

        """
        really_close = super().exit
        if self.tabs:

            def final_cb() -> None:
                """Really exit after the last tab in the chain is closed."""
                self.cleanup_closed_tab(self.tabs[0])
                really_close(*args, **kwargs)

            def create_cb(close_tab: Tab, cleanup_tab: Tab, cb: Callable) -> Callable:
                """Generate a tab close chaining callbacks.

                Cleans up after the previously closed tab, and requests to close the
                next tab in the chain.

                Args:
                    close_tab: The tab to close
                    cleanup_tab: The previously closed tab to cleanup
                    cb: The callback to call when work is complete

                Returns:
                    A callback function which cleans up `cleanup_tab` and closes
                        `close_tab`.

                """

                def inner() -> None:
                    self.cleanup_closed_tab(cleanup_tab)
                    close_tab.close(cb=cb)

                return inner

            cb = final_cb
            for close_tab, cleanup_tab in zip(self.tabs, self.tabs[1:]):
                cb = create_cb(close_tab, cleanup_tab, cb)
            self.tabs[-1].close(cb)
        else:
            really_close(*args, **kwargs)

    @property
    def notebook(self) -> Notebook | None:
        """Return the currently active notebook."""
        if isinstance(self.tab, Notebook):
            return self.tab
        return None

    @property
    def cell(self) -> Cell | None:
        """Return the currently active cell."""
        if isinstance(self.tab, Notebook):
            return self.tab.cell
        return None

    def load_menu_items(self) -> list[MenuItem]:
        """Load the list of menu items to display in the menu."""
        separator = MenuItem(separator=True)
        return [
            MenuItem(
                "File",
                children=[
                    get_cmd("new-notebook").menu,
                    get_cmd("open-file").menu,
                    separator,
                    get_cmd("save-file").menu,
                    get_cmd("save-as").menu,
                    get_cmd("close-tab").menu,
                    separator,
                    get_cmd("quit").menu,
                ],
            ),
            MenuItem(
                "Edit",
                children=[
                    get_cmd("undelete-cells").menu,
                    separator,
                    get_cmd("cut-cells").menu,
                    get_cmd("copy-cells").menu,
                    get_cmd("paste-cells").menu,
                    separator,
                    get_cmd("copy-outputs").menu,
                    get_cmd("clear-cell-outputs").menu,
                    separator,
                    get_cmd("find").menu,
                    separator,
                    get_cmd("reformat-cells").menu,
                    get_cmd("reformat-notebook").menu,
                ],
            ),
            MenuItem(
                "Run",
                children=[
                    get_cmd("run-selected-cells").menu,
                    get_cmd("run-all-cells").menu,
                ],
            ),
            MenuItem(
                "Kernel",
                children=[
                    get_cmd("interrupt-kernel").menu,
                    get_cmd("restart-kernel").menu,
                    get_cmd("change-kernel").menu,
                ],
            ),
            MenuItem(
                "Tabs",
                children=[
                    get_cmd("next-tab").menu,
                    get_cmd("previous-tab").menu,
                    separator,
                    MenuItem(
                        "Tab mode",
                        children=[
                            get_cmd(f"set-tab-mode-{choice}").menu
                            for choice in self.config.get_item("tab_mode").choices
                        ],
                    ),
                ],
            ),
            MenuItem(
                "Settings",
                children=[
                    self.config.get_item("edit_mode").menu,
                    self.config.get_item("set_cursor_shape").menu,
                    separator,
                    self.config.get_item("color_scheme").menu,
                    self.config.get_item("syntax_theme").menu,
                    get_cmd("toggle-background-pattern").menu,
                    get_cmd("toggle-show-cell-borders").menu,
                    get_cmd("toggle-tmux-graphics").menu,
                    separator,
                    get_cmd("toggle-expand").menu,
                    get_cmd("toggle-line-numbers").menu,
                    MenuItem(
                        "UI Elements",
                        children=[
                            get_cmd("toggle-always-show-tab-bar").menu,
                            get_cmd("toggle-show-side-bar").menu,
                            get_cmd("toggle-show-status-bar").menu,
                            get_cmd("toggle-show-scroll-bar").menu,
                        ],
                    ),
                    separator,
                    MenuItem(
                        "Cell formatting",
                        children=[
                            get_cmd("toggle-autoformat").menu,
                            separator,
                            get_cmd("toggle-format-black").menu,
                            get_cmd("toggle-format-isort").menu,
                            get_cmd("toggle-format-ssort").menu,
                        ],
                    ),
                    get_cmd("toggle-autocomplete").menu,
                    get_cmd("toggle-autosuggest").menu,
                    get_cmd("toggle-autoinspect").menu,
                    get_cmd("toggle-run-after-external-edit").menu,
                ],
            ),
            MenuItem(
                "Help",
                children=[
                    get_cmd("show-command-palette").menu,
                    get_cmd("keyboard-shortcuts").menu,
                    get_cmd("view-documentation").menu,
                    separator,
                    get_cmd("view-logs").menu,
                    separator,
                    get_cmd("about").menu,
                ],
            ),
        ]

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd()
    def _new_notebook() -> None:
        """Create a new file."""
        from euporie.notebook.current import get_app

        app = get_app()
        app.tabs.append(Notebook(app, None))
        app.tabs[-1].focus()

    @staticmethod
    @add_cmd()
    def _view_documentation() -> None:
        """Open the documentation in a web-view tab."""
        from euporie.core.current import get_app

        get_app().open_file(UPath("https://euporie.readthedocs.io/"))

    # ################################### Settings ####################################

    add_setting(
        name="tab_mode",
        flags=["--tab-mode"],
        type_=str,
        choices=[mode.value for mode in TabMode],
        default="stack",
        help_="The method used to display multiple tabs",
        description="""
            Determines how multiple tabs are displayed when more than one tab is open.
            * ``stack`` displays one tab at a time with a tab-bar
            * ``tile_horizontally`` displays tabs side-by-side
            * ``tile_vertically`` displays tabs one-atop-the-next
        """,
    )

    add_setting(
        name="always_show_tab_bar",
        flags=["--always-show-tab-bar"],
        type_=bool,
        help_="Always show the tab bar",
        default=False,
        description="""
            When set, the tab bar will always be shown - otherwise the tab bar is only
            shown when multiple tabs are open.
        """,
    )

    add_setting(
        name="background_pattern",
        flags=["--background-pattern", "--bg-pattern"],
        type_=int,
        choices=list(range(6)),
        help_="The background pattern to use",
        default=2,
        schema={
            "minimum": 0,
            "maximum": 5,
        },
        description="""
            The background pattern to use when the notebook is narrower than the
            available width. Zero mean no pattern is used.
        """,
    )

    add_setting(
        name="background_character",
        flags=["--background-character", "--bg-char"],
        type_=str,
        help_="Character for background pattern",
        default="·",
        schema={
            "maxLength": 1,
        },
        description="""
            The character to use when drawing the background pattern.

            Recommended characters include: "·", "⬤", "╳", "╱", "╲", "░", "▒", "▓", "▞", "╬"
        """,
    )

    add_setting(
        name="run_after_external_edit",
        flags=["--run-after-external-edit"],
        type_=bool,
        help_="Run cells after editing externally",
        default=False,
        description="""
            Whether to execute a cell immediately after editing in `$EDITOR`.
        """,
    )

    add_setting(
        name="run",
        flags=["--run"],
        type_=bool,
        help_="Run the notebook files when loaded",
        default=False,
        description="""
            If set, notebooks will be run automatically when opened, or if previewing a
            file, the notebooks will be run before being output.
        """,
    )

    add_setting(
        name="show_top_bar",
        flags=["--show-top-bar"],
        type_=bool,
        title="top bar",
        help_="Show the top bar",
        default=True,
        schema={"type": "boolean"},
        description="""
            Whether the top bar should be shown at the top of the screen.
        """,
    )

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.notebook.app.NotebookApp": {
                "new-notebook": "c-n",
                "toggle-show-top-bar": ("A-m"),
            }
        }
    )



"""Contain functions which convert data to base64 format."""

from __future__ import annotations

import base64
from typing import TYPE_CHECKING

from euporie.core.convert.core import register

if TYPE_CHECKING:
    from pathlib import Path


@register(
    from_="png",
    to="base64-png",
    filter_=True,
)
@register(
    from_="jpeg",
    to="base64-jpeg",
    filter_=True,
)
@register(
    from_="pdf",
    to="base64-pdf",
    filter_=True,
)
@register(
    from_="svg",
    to="base64-svg",
    filter_=True,
)
async def bytes_to_base64_py(
    data: str | bytes,
    width: int | None = None,
    height: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
    initial_format: str = "",
) -> str:
    """Convert bytes to base64 encoded data."""
    if isinstance(data, str):
        data = data.encode()
    return base64.b64encode(data).decode().replace("\n", "").strip()



"""Defines a web-view control."""

from __future__ import annotations

import logging
import weakref
from typing import TYPE_CHECKING, cast

from prompt_toolkit.cache import FastDictCache
from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters import Condition
from prompt_toolkit.formatted_text.utils import split_lines
from prompt_toolkit.layout.containers import Float, Window
from prompt_toolkit.layout.controls import UIContent, UIControl
from prompt_toolkit.layout.screen import WritePosition
from prompt_toolkit.mouse_events import MouseButton, MouseEvent, MouseEventType
from prompt_toolkit.utils import Event
from upath import UPath

from euporie.core.commands import add_cmd
from euporie.core.convert.core import convert, get_format
from euporie.core.current import get_app
from euporie.core.data_structures import DiInt
from euporie.core.ft.html import HTML, Node
from euporie.core.ft.utils import fragment_list_width, paste
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.path import parse_path
from euporie.core.utils import run_in_thread_with_context
from euporie.core.widgets.display import (
    GraphicWindow,
    NotVisible,
    select_graphic_control,
)
from euporie.core.widgets.page import BoundedWritePosition

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Callable, Iterable

    from prompt_toolkit.formatted_text.base import AnyFormattedText, StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import (
        KeyBindingsBase,
        NotImplementedOrNone,
    )
    from prompt_toolkit.layout.controls import GetLinePrefixCallable
    from prompt_toolkit.layout.mouse_handlers import MouseHandler
    from prompt_toolkit.layout.screen import Screen


log = logging.getLogger(__name__)


@Condition
def webview_has_focus() -> bool:
    """Determine if there is a currently focused webview."""
    return isinstance(get_app().layout.current_control, WebViewControl)


class WebViewControl(UIControl):
    """Web view displays.

    A control which displays rendered HTML content.
    """

    _window: Window

    def __init__(
        self,
        url: str | Path = "about:blank",
        link_handler: Callable | None = None,
    ) -> None:
        """Create a new web-view control instance."""
        self._cursor_position = Point(0, 0)
        self.loading = False
        self.resizing = False
        self.rendering = False
        self.lines: list[StyleAndTextTuples] = []
        self.graphic_positions: dict[str, tuple[int, int]] = {}
        self.width = 0
        self.height = 0
        self.url: Path = UPath(url)
        self.status: list[AnyFormattedText] = []
        self.link_handler = link_handler or self.load_url

        self.rendered = Event(self)
        self.on_cursor_position_changed = Event(self)

        self.prev_stack: list[Path] = []
        self.next_stack: list[Path] = []

        # self.cursor_processor = CursorProcessor(
        #     lambda: self.cursor_position, style="fg:red"
        # )

        self.key_bindings = load_registered_bindings(
            "euporie.web.widgets.webview.WebViewControl"
        )

        self._dom_cache: FastDictCache[tuple[Path], HTML] = FastDictCache(
            get_value=self.get_dom, size=100
        )
        self._line_cache: FastDictCache[
            tuple[HTML, int, int], list[StyleAndTextTuples]
        ] = FastDictCache(get_value=self.get_lines, size=100_000)
        self._content_cache: FastDictCache = FastDictCache(self.get_content, size=1_000)
        self._graphic_float_cache: FastDictCache = FastDictCache(
            self.get_graphic_float, size=1_000
        )

        self.load_url(url)

    @property
    def window(self) -> Window:
        """Get the control's window."""
        try:
            return self._window
        except AttributeError:
            for window in get_app().layout.find_all_windows():
                if window.content == self:
                    self._window = window
                    return window
        return Window()

    @property
    def cursor_position(self) -> Point:
        """Get the cursor position."""
        return self._cursor_position

    @cursor_position.setter
    def cursor_position(self, value: Point) -> None:
        """Set the cursor position."""
        changed = self._cursor_position != value
        self._cursor_position = value
        if changed:
            self.on_cursor_position_changed.fire()

    def get_dom(self, url: Path) -> HTML:
        """Load a HTML page as renderable formatted text."""
        markup = str(
            convert(
                data=url.read_text(),
                from_=(format_ := get_format(url, default="html")),
                to="html",
                path=url,
            )
        )
        return HTML(
            markup=markup,
            base=url,
            mouse_handler=self._node_mouse_handler,
            paste_fixed=False,
            _initial_format=format_,
        )

    @property
    def dom(self) -> HTML:
        """Return the dom for the current URL."""
        return self._dom_cache[self.url,]

    @property
    def title(self) -> str:
        """Return the title of the current HTML page."""
        if self.loading:
            return self.url.name
        if dom := self.dom:
            return dom.title
        return ""

    def get_lines(self, dom: HTML, width: int, height: int) -> list[StyleAndTextTuples]:
        """Render a HTML page as lines of formatted text."""
        app = get_app()

        lines = list(split_lines(dom.render(width, height)))

        # Check for graphics
        self.graphic_positions.clear()
        for y, line in enumerate(lines):
            x = 0
            for style, text, *_ in line:
                for token in style.split():
                    token = token[1:-1]
                    if token.startswith("Image_"):
                        self.graphic_positions[token] = (x, y)
                        # Get graphic float for this image and update its position
                        graphic_float = self._graphic_float_cache[dom, token]
                        # Register graphic with application
                        if graphic_float:
                            app.graphics.add(graphic_float)
                        break
                if "[ZeroWidthEscape]" not in style:
                    x += len(text)

        # Check for floating graphics and create graphics (do not position them yet)
        for line in split_lines(dom.fixed_mask):
            for style, *_ in line:
                for token in style.split():
                    token = token[1:-1]
                    if token.startswith("Image_"):
                        # Get graphic float for this image and update its position
                        graphic_float = self._graphic_float_cache[dom, token]
                        # Register graphic with application
                        if graphic_float:
                            app.graphics.add(graphic_float)

        return lines

    def load_url(self, url: str | Path, **kwargs: Any) -> None:
        """Load a new URL."""
        save_to_history = kwargs.get("save_to_history", True)
        # Trigger "loading" view
        self.loading = True
        # Clear graphics
        self.graphic_positions.clear()
        # Update navigation history
        if self.url and save_to_history:
            self.prev_stack.append(self.url)
            self.next_stack.clear()
        # Update url
        self.url = UPath(url)
        # Scroll to the top
        self.window.vertical_scroll = 0
        self.cursor_position = Point(0, 0)
        # Signal that the webview has updated
        self.rendered.fire()

    def nav_prev(self) -> None:
        """Navigate forwards through the browser history."""
        if self.url and self.prev_stack:
            self.next_stack.append(self.url)
            self.load_url(self.prev_stack.pop(), save_to_history=False)

    def nav_next(self) -> None:
        """Navigate backwards through the browser history."""
        if self.url and self.next_stack:
            self.prev_stack.append(self.url)
            self.load_url(self.next_stack.pop(), save_to_history=False)

    def render(self) -> None:
        """Render the HTML DOM in a thread."""

        def _render() -> None:
            assert self.url is not None
            # Potentially redirect url
            self.url = parse_path(self.url)
            self.lines = self._line_cache[self.dom, self.width, self.height]
            self.loading = False
            self.resizing = False
            self.rendering = False
            self.rendered.fire()

        if not self.rendering:
            self.rendering = True
            run_in_thread_with_context(_render)

    def reset(self) -> None:
        """Reset the state of the control."""

    def preferred_width(self, max_available_width: int) -> int | None:
        """Calculate and return the preferred width of the control."""
        return None

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix: GetLinePrefixCallable | None,
    ) -> int | None:
        """Calculate and return the preferred height of the control."""
        return None

    def is_focusable(self) -> bool:
        """Tell whether this user control is focusable."""
        return True

    def get_graphic_float(self, dom: HTML, key: str) -> Float | None:
        """Create a graphical float for an image."""
        graphic_info = dom.graphic_info.get(key)
        if graphic_info is None:
            return None

        GraphicControl = select_graphic_control(format_=graphic_info["format_"])
        if GraphicControl is None:
            return None

        # TODO - cache this
        def get_position(screen: Screen) -> tuple[WritePosition, DiInt]:
            """Get the position and bbox of a graphic."""
            if key not in self.graphic_positions:
                raise NotVisible

            # Hide graphic if webview is not in layout
            if screen.visible_windows_to_write_positions.get(self.window) is None:
                raise NotVisible

            graphic_info = dom.graphic_info.get(key)
            if graphic_info is None:
                raise NotVisible

            render_info = self.window.render_info
            if render_info is None:
                raise NotVisible

            x, y = self.graphic_positions[key]

            content_width = max(0, graphic_info["cols"])
            horizontal_scroll = getattr(render_info, "horizontal_scroll", 0)

            if horizontal_scroll >= x + content_width:
                raise NotVisible

            content_height = max(0, graphic_info["rows"])
            vertical_scroll = render_info.vertical_scroll

            if vertical_scroll >= y + content_height:
                raise NotVisible

            x_offset = render_info._x_offset
            y_offset = render_info._y_offset

            xpos = max(x_offset, x - horizontal_scroll + x_offset)
            if xpos >= x_offset + render_info.window_width:
                raise NotVisible
            ypos = max(y_offset, y - vertical_scroll + y_offset)
            if ypos >= y_offset + render_info.window_height:
                raise NotVisible

            bbox = DiInt(
                top=max(0, render_info.vertical_scroll - y),
                right=0,  # TODO
                bottom=max(
                    0, (ypos + content_height) - (y_offset + render_info.window_height)
                ),
                left=0,  # TODO
            )
            if (width := content_width - bbox.left - bbox.right) < 1:
                raise NotVisible
            if (height := content_height - bbox.top - bbox.bottom) < 1:
                raise NotVisible

            write_position = BoundedWritePosition(
                xpos=xpos,
                ypos=ypos,
                width=width,
                height=height,
                bbox=bbox,
            )

            return write_position, bbox

        def _sizing_func() -> tuple[int, float]:
            graphic_info = dom.graphic_info[key]
            return (
                graphic_info["cols"],
                graphic_info["aspect"],
            )

        bg_color = graphic_info["bg"]
        graphic_float = Float(
            graphic_window := GraphicWindow(
                content=(
                    graphic_control := GraphicControl(
                        graphic_info["data"],
                        format_=graphic_info["format_"],
                        path=graphic_info["path"],
                        fg_color=graphic_info["fg"],
                        bg_color=graphic_info["bg"],
                        sizing_func=_sizing_func,
                    )
                ),
                position=get_position,
                style=f"bg:{bg_color}" if bg_color else "",
            ),
        )
        # Hide the graphic if the float is deleted
        weak_float_ref = weakref.ref(graphic_float)
        graphic_window.filter &= Condition(
            lambda: weak_float_ref() in get_app().graphics
        )
        # Hide the graphic if the float is deleted
        weakref.finalize(graphic_float, graphic_control.close)

        return graphic_float

    def get_content(
        self,
        url: Path,
        loading: bool,
        resizing: bool,
        width: int,
        # height: int,
        cursor_position: Point,
    ) -> UIContent:
        """Create a cacheable UIContent."""
        if self.loading:
            lines = [
                cast("StyleAndTextTuples", []),
                cast(
                    "StyleAndTextTuples",
                    [("", " " * ((width - 8) // 2)), ("class:loading", "Loading…")],
                ),
            ]
        else:
            lines = self.lines

        def get_line(i: int) -> StyleAndTextTuples:
            try:
                line = lines[i]
            except IndexError:
                return []

            # Overlay fixed lines onto this line
            if not loading:
                dom = self._dom_cache[url,]
                if dom.fixed:
                    visible_line = max(0, i - self.window.vertical_scroll)
                    fixed_lines = list(split_lines(dom.fixed_mask))
                    if visible_line < len(fixed_lines):
                        # Paste the fixed line over the current line
                        fixed_line = fixed_lines[visible_line]
                        line = paste(
                            fixed_lines[visible_line], line, 0, 0, transparent=True
                        )
                        # Update graphic positions on the fixed line
                        x = 0
                        for style, text, *_ in fixed_line:
                            for token in style.split():
                                token = token[1:-1]
                                if token.startswith("Image_"):
                                    self.graphic_positions[token] = (x, i)
                                    break
                            if "[ZeroWidthEscape]" not in style:
                                x += len(text)

                # Apply processors
                # merged_processor = self.cursor_processor
                # line = lines[i]
                # transformation = merged_processor.apply_transformation(
                #     TransformationInput(
                #         buffer_control=self, document=Document(), lineno=i, source_to_display=lambda i: i, fragments=line, width=width, height=height,
                #     )
                # )
                # return transformation.fragments

            return line

        return UIContent(
            get_line=get_line,
            line_count=len(lines),
            cursor_position=self.cursor_position,
            show_cursor=False,
        )

    def create_content(self, width: int, height: int) -> UIContent:
        """Generate the content for this user control.

        Returns:
            A :class:`.UIContent` instance.
        """
        # Trigger a re-render if things have changed
        if self.loading:
            self.render()
        if width != self.width:  # or height != self.height:
            self.resizing = True
            self.width = width
            self.height = height
            self.render()

        return self._content_cache[
            self.url,
            self.loading,
            False,  # self.resizing,
            width,
            # height,
            self.cursor_position,
        ]

    def _node_mouse_handler(
        self, node: Node, mouse_event: MouseEvent
    ) -> NotImplementedOrNone:
        """Handle click events."""
        url = node.attrs.get("_link_path")
        title = node.attrs.get("title")
        if url:
            # TODO - Check for #anchor links and scroll accordingly
            if (
                mouse_event.button == MouseButton.LEFT
                and mouse_event.event_type == MouseEventType.MOUSE_UP
            ):
                self.link_handler(url, save_to_history=True, new_tab=False)
                return None
            elif (
                mouse_event.button == MouseButton.MIDDLE
                and mouse_event.event_type == MouseEventType.MOUSE_UP
            ):
                self.link_handler(url, save_to_history=False, new_tab=True)
                return None
        if url or title:
            if mouse_event.event_type == MouseEventType.MOUSE_MOVE:
                self.status.clear()
                if title:
                    self.status.append(str(title))
                if url:
                    self.status.append(str(url))
                return None
        return NotImplemented

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Handle mouse events.

        When `NotImplemented` is returned, it means that the given event is not
        handled by the `UIControl` itself. The `Window` or key bindings can
        decide to handle this event as scrolling or changing focus.

        Args:
            mouse_event: `MouseEvent` instance.

        Returns:
            NotImplemented if the UI does not need to be updates, None if it does
        """
        # Focus on mouse down
        if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
            if not (layout := get_app().layout).has_focus(self):
                layout.focus(self)
                return None
            return NotImplemented
        # if mouse_event.event_type == MouseEventType.MOUSE_MOVE:
        # self.cursor_position = mouse_event.position

        # if mouse_event.event_type == MouseEventType.MOUSE_UP:
        handler: MouseHandler | None = None
        try:
            content = self._content_cache[
                self.url,
                self.loading,
                self.resizing,
                self.width,
                # self.height,
                self.cursor_position,
            ]
            line = content.get_line(mouse_event.position.y)
        except IndexError:
            return NotImplemented
        else:
            # Find position in the fragment list.
            xpos = mouse_event.position.x
            # Find mouse handler for this character.
            count = 0
            for item in line:
                count += len(item[1])
                if count > xpos:
                    if len(item) >= 3:
                        handler = item[2]
                        if callable(handler):
                            return handler(mouse_event)
                    else:
                        break

        if callable(handler):
            return handler

        if self.status:
            self.status.clear()
            return None

        return NotImplemented

    @property
    def content_width(self) -> int:
        """Return the width of the content."""
        return max(fragment_list_width(line) for line in self.lines)

    def move_cursor_down(self) -> None:
        """Move the cursor down one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x, y=y + 1)

    def move_cursor_up(self) -> None:
        """Move the cursor up one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x, y=max(0, y - 1))

    def move_cursor_left(self) -> None:
        """Move the cursor down one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=max(0, x - 1), y=y)

    def move_cursor_right(self) -> None:
        """Move the cursor up one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x + 1, y=y)

    def get_key_bindings(self) -> KeyBindingsBase | None:
        """Return key bindings that are specific for this user control.

        Returns:
            A :class:`.KeyBindings` object if some key bindings are specified, or
                `None` otherwise.
        """
        return self.key_bindings

    def get_invalidate_events(self) -> Iterable[Event[object]]:
        """Return a list of `Event` objects, which can be a generator.

        the application collects all these events, in order to bind redraw
        handlers to these events.
        """
        yield self.rendered
        yield self.on_cursor_position_changed
        if dom := self.dom:
            yield dom.on_update

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _webview_nav_prev() -> None:
        """Navigate backwards in the browser history."""
        from euporie.web.widgets.webview import WebViewControl

        current_control = get_app().layout.current_control
        if isinstance(current_control, WebViewControl):
            current_control.nav_prev()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _webview_nav_next() -> None:
        """Navigate forwards in the browser history."""
        from euporie.web.widgets.webview import WebViewControl

        current_control = get_app().layout.current_control
        if isinstance(current_control, WebViewControl):
            current_control.nav_next()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _scroll_webview_left() -> None:
        """Scroll the display up one line."""
        from euporie.core.widgets.display import DisplayWindow

        window = get_app().layout.current_window
        assert isinstance(window, DisplayWindow)
        window._scroll_left()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _scroll_webview_right() -> None:
        """Scroll the display down one line."""
        from euporie.core.widgets.display import DisplayWindow

        window = get_app().layout.current_window
        assert isinstance(window, DisplayWindow)
        window._scroll_right()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _scroll_webview_up() -> None:
        """Scroll the display up one line."""
        get_app().layout.current_window._scroll_up()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _scroll_webview_down() -> None:
        """Scroll the display down one line."""
        get_app().layout.current_window._scroll_down()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _page_up_webview() -> None:
        """Scroll the display up one page."""
        window = get_app().layout.current_window
        if window.render_info is not None:
            for _ in range(window.render_info.window_height):
                window._scroll_up()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _page_down_webview() -> None:
        """Scroll the display down one page."""
        window = get_app().layout.current_window
        if window.render_info is not None:
            for _ in range(window.render_info.window_height):
                window._scroll_down()

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _go_to_start_of_webview() -> None:
        """Scroll the display to the top."""
        from euporie.web.widgets.webview import WebViewControl

        current_control = get_app().layout.current_control
        if isinstance(current_control, WebViewControl):
            current_control.cursor_position = Point(0, 0)

    @staticmethod
    @add_cmd(filter=webview_has_focus)
    def _go_to_end_of_webview() -> None:
        """Scroll the display down one page."""
        from euporie.web.widgets.webview import WebViewControl

        layout = get_app().layout
        current_control = layout.current_control
        window = layout.current_window
        if (
            isinstance(current_control, WebViewControl)
            and window.render_info is not None
        ):
            current_control.cursor_position = Point(
                0, window.render_info.ui_content.line_count - 1
            )

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.web.widgets.webview.WebViewControl": {
                "scroll-webview-left": "left",
                "scroll-webview-right": "right",
                "scroll-webview-up": ["up", "k"],
                "scroll-webview-down": ["down", "j"],
                "page-up-webview": "pageup",
                "page-down-webview": "pagedown",
                "go-to-start-of-webview": "home",
                "go-to-end-of-webview": "end",
                "webview-nav-prev": ("A-left"),
                "webview-nav-next": ("A-right"),
            }
        }
    )


if __name__ == "__main__":
    import sys

    from prompt_toolkit.application.application import Application
    from prompt_toolkit.key_binding.key_bindings import KeyBindings
    from prompt_toolkit.layout.layout import Layout
    from prompt_toolkit.output.color_depth import ColorDepth
    from prompt_toolkit.styles.style import Style

    from euporie.core.style import HTML_STYLE
    from euporie.core.widgets.display import DisplayWindow
    from euporie.web.widgets.webview import WebViewControl  # noqa F811

    kb = KeyBindings()
    kb.add("q")(lambda event: event.app.exit())
    layout = Layout(container=DisplayWindow(WebViewControl(UPath(sys.argv[-1]))))
    app: Application = Application(
        layout=layout,
        key_bindings=kb,
        full_screen=True,
        style=Style(HTML_STYLE),
        mouse_support=True,
        color_depth=ColorDepth.DEPTH_24_BIT,
    )
    app.run()



"""Define editor key-bindings and commands for input completions."""

from __future__ import annotations

import logging

from prompt_toolkit.application import get_app
from prompt_toolkit.completion import Completion
from prompt_toolkit.filters import (
    buffer_has_focus,
    completion_is_selected,
    has_completions,
    has_selection,
)
from prompt_toolkit.key_binding.bindings.named_commands import (
    menu_complete,
    menu_complete_backward,
)

from euporie.core.commands import add_cmd
from euporie.core.filters import buffer_is_code, cursor_in_leading_ws, insert_mode
from euporie.core.key_binding.registry import register_bindings

log = logging.getLogger(__name__)


add_cmd(
    filter=buffer_has_focus
    & insert_mode
    & ~has_selection
    & buffer_is_code
    & ~cursor_in_leading_ws,
    hidden=True,
    name="next-completion",
    description="Show the completion menu and select the next completion.",
)(menu_complete)


add_cmd(
    filter=buffer_has_focus
    & completion_is_selected
    & insert_mode
    & ~has_selection
    & ~cursor_in_leading_ws,
    hidden=True,
    name="previous-completion",
    description="Show the completion menu and select the previous completion.",
)(menu_complete_backward)


@add_cmd(
    filter=has_completions,
    hidden=True,
    eager=True,
)
def cancel_completion() -> None:
    """Cancel a completion."""
    get_app().current_buffer.cancel_completion()


@add_cmd(filter=completion_is_selected, hidden=True)
def accept_completion() -> None:
    """Accept a selected completion."""
    buffer = get_app().current_buffer
    complete_state = buffer.complete_state
    if complete_state:
        if isinstance(complete_state.current_completion, Completion):
            buffer.apply_completion(complete_state.current_completion)
            get_app().layout.focus(buffer)


register_bindings(
    {
        "euporie.core.app.BaseApp": {
            "next-completion": "c-i",
            "previous-completion": "s-tab",
            "cancel-completion": "escape",
            "accept-completion": "enter",
        }
    }
)



"""Contain the main Application class which runs euporie.core."""

from __future__ import annotations

import asyncio
import json
import logging
import signal
import sys
from functools import partial
from pathlib import PurePath
from typing import TYPE_CHECKING, cast
from weakref import WeakSet

from prompt_toolkit.application.application import Application, _CombinedRegistry
from prompt_toolkit.application.current import create_app_session
from prompt_toolkit.clipboard import InMemoryClipboard
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
from prompt_toolkit.cursor_shapes import CursorShape, CursorShapeConfig
from prompt_toolkit.data_structures import Point
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters import Condition, buffer_has_focus, to_filter
from prompt_toolkit.input.defaults import create_input
from prompt_toolkit.key_binding.bindings.basic import (
    load_basic_bindings as load_ptk_basic_bindings,
)
from prompt_toolkit.key_binding.bindings.cpr import load_cpr_bindings
from prompt_toolkit.key_binding.bindings.emacs import (
    load_emacs_bindings,
    load_emacs_search_bindings,
    load_emacs_shift_selection_bindings,
)
from prompt_toolkit.key_binding.bindings.mouse import (
    load_mouse_bindings as load_ptk_mouse_bindings,
)
from prompt_toolkit.key_binding.bindings.vi import (
    load_vi_bindings,
    load_vi_search_bindings,
)
from prompt_toolkit.key_binding.key_bindings import (
    ConditionalKeyBindings,
    merge_key_bindings,
)
from prompt_toolkit.layout.containers import Float, FloatContainer, Window, to_container
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.output import ColorDepth
from prompt_toolkit.output.defaults import create_output
from prompt_toolkit.output.vt100 import Vt100_Output as PtkVt100_Output
from prompt_toolkit.styles import (
    BaseStyle,
    ConditionalStyleTransformation,
    DummyStyle,
    SetDefaultColorStyleTransformation,
    Style,
    SwapLightAndDarkStyleTransformation,
    merge_style_transformations,
    merge_styles,
    style_from_pygments_cls,
)
from pygments.styles import STYLE_MAP as pygments_styles
from pygments.styles import get_style_by_name
from pyperclip import determine_clipboard
from upath import UPath

from euporie.core.commands import add_cmd
from euporie.core.config import Config, add_setting
from euporie.core.convert.core import get_mime
from euporie.core.current import get_app
from euporie.core.filters import in_tmux, insert_mode, replace_mode, tab_has_focus
from euporie.core.io import Vt100_Output, Vt100Parser
from euporie.core.key_binding.key_processor import KeyProcessor
from euporie.core.key_binding.micro_state import MicroState
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.key_binding.vi_state import ViState
from euporie.core.log import setup_logs
from euporie.core.path import parse_path
from euporie.core.renderer import Renderer
from euporie.core.style import (
    DEFAULT_COLORS,
    HTML_STYLE,
    IPYWIDGET_STYLE,
    LOG_STYLE,
    MIME_STYLE,
    ColorPalette,
    build_style,
)
from euporie.core.terminal import TerminalInfo
from euporie.core.utils import ChainedList
from euporie.core.widgets.decor import Shadow
from euporie.core.widgets.menu import CompletionsMenu

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from pathlib import Path
    from types import FrameType
    from typing import Any, Callable, TypeVar

    # from prompt_toolkit.application import _AppResult
    from prompt_toolkit.clipboard import Clipboard
    from prompt_toolkit.contrib.ssh import PromptToolkitSSHSession
    from prompt_toolkit.filters import Filter, FilterOrBool
    from prompt_toolkit.input import Input
    from prompt_toolkit.layout.layout import FocusableElement
    from prompt_toolkit.layout.screen import WritePosition
    from prompt_toolkit.output import Output

    from euporie.core.config import Setting
    from euporie.core.tabs.base import Tab
    from euporie.core.terminal import TerminalQuery
    from euporie.core.widgets.dialog import Dialog
    from euporie.core.widgets.pager import Pager
    from euporie.core.widgets.search_bar import SearchBar

    _AppResult = TypeVar("_AppResult")

log = logging.getLogger(__name__)


_COLOR_DEPTHS = {
    1: ColorDepth.DEPTH_1_BIT,
    4: ColorDepth.DEPTH_4_BIT,
    8: ColorDepth.DEPTH_8_BIT,
    24: ColorDepth.DEPTH_24_BIT,
}


class CursorConfig(CursorShapeConfig):
    """Determine which cursor mode to use."""

    def get_cursor_shape(self, app: Application[Any]) -> CursorShape:
        """Return the cursor shape to be used in the current state."""
        if isinstance(app, BaseApp) and app.config.set_cursor_shape:
            if insert_mode():
                if app.config.cursor_blink:
                    return CursorShape.BLINKING_BEAM
                else:
                    return CursorShape.BEAM
            elif replace_mode():
                if app.config.cursor_blink:
                    return CursorShape.BLINKING_UNDERLINE
                else:
                    return CursorShape.UNDERLINE
        return CursorShape.BLOCK

    # ################################### Settings ####################################w

    add_setting(
        name="set_cursor_shape",
        flags=["--set-cursor-shape"],
        type_=bool,
        default=True,
        menu_title="Change cursor shape",
        help_="Whether to set the shape of the cursor depending on the editing mode",
        description="""
            When set to True, the euporie will set the shape of the terminal's cursor
            to a beam in insert mode and and underline in replace mode when editing.
    """,
    )
    add_setting(
        name="cursor_blink",
        flags=["--cursor-blink"],
        type_=bool,
        default=False,
        help_="Whether to blink the cursor",
        description="""
            When set to True, the cursor will blink.
    """,
    )


class BaseApp(Application):
    """All euporie apps.

    The base euporie application class.

    This subclasses the `prompt_toolkit.application.Application` class, so application
    wide methods can be easily added.
    """

    name: str
    color_palette: ColorPalette
    mouse_position: Point

    config = Config()
    need_mouse_support: bool = False
    log_stdout_level: str = "CRITICAL"

    def __init__(
        self,
        title: str | None = None,
        set_title: bool = True,
        leave_graphics: FilterOrBool = True,
        extend_renderer_height: FilterOrBool = False,
        extend_renderer_width: FilterOrBool = False,
        **kwargs: Any,
    ) -> None:
        """Instantiate euporie specific application variables.

        After euporie specific application variables are instantiated, the application
        instance is initiated.

        Args:
            title: The title string to set in the terminal
            set_title: Whether to set the terminal title
            leave_graphics: A filter which determines if graphics should be cleared
                from the display when they are no longer active
            extend_renderer_height: Whether the renderer height should be extended
                beyond the height of the display
            extend_renderer_width: Whether the renderer width should be extended
                beyond the height of the display
            kwargs: The key-word arguments for the :py:class:`Application`

        """
        self.loaded = False
        # Initialise the application
        super().__init__(
            **{
                **{
                    "color_depth": self.config.color_depth,
                    "editing_mode": self.get_edit_mode(),
                    "mouse_support": Condition(lambda: self.need_mouse_support),
                    "cursor": CursorConfig(),
                },
                **kwargs,
            }
        )
        # Use custom renderer
        self.renderer = Renderer(
            self._merged_style,
            self.output,
            full_screen=self.full_screen,
            mouse_support=self.mouse_support,
            cpr_not_supported_callback=self.cpr_not_supported_callback,
            extend_height=extend_renderer_height,
            extend_width=extend_renderer_width,
        )
        # Contains the opened tab containers
        self.tabs: list[Tab] = []
        # Holds the search bar to pass to cell inputs
        self.search_bar: SearchBar | None = None
        # Holds the index of the current tab
        self._tab_idx = 0
        # Add state for micro key-bindings
        self.micro_state = MicroState()
        # Load the terminal information system
        self.term_info = TerminalInfo(self.input, self.output, self.config)
        # Floats at the app level
        self.leave_graphics = to_filter(leave_graphics)
        self.graphics: WeakSet[Float] = WeakSet()
        self.dialogs: dict[str, Dialog] = {}
        self.menus: dict[str, Float] = {}
        self.floats = ChainedList(
            self.graphics,
            self.dialogs.values(),
            self.menus.values(),
        )
        # Continue loading when the application has been launched
        # and an event loop has been created
        self.pre_run_callables = [self.pre_run]
        self.post_load_callables: list[Callable[[], None]] = []
        # Set default vi input mode to navigate
        self.vi_state = ViState()
        # Set a long timeout for mappings (e.g. dd)
        self.timeoutlen = 1.0
        # Set a short timeout for flushing input
        self.ttimeoutlen = 0.0
        # Use a custom key-processor which does not wait after escape keys
        self.key_processor = KeyProcessor(_CombinedRegistry(self))
        # List of key-bindings groups to load
        self.bindings_to_load = ["euporie.core.app.BaseApp"]
        # Determines which clipboard mechanism to use
        self.clipboard: Clipboard = (
            PyperclipClipboard() if determine_clipboard()[0] else InMemoryClipboard()
        )
        # Allow hiding element when manually redrawing app
        self._redrawing = False
        self.redrawing = Condition(lambda: self._redrawing)
        # Add an optional pager
        self.pager: Pager | None = None
        # Stores the initially focused element
        self.focused_element: FocusableElement | None = None
        # Set the terminal title
        self.set_title = to_filter(set_title)
        self.title = title or self.__class__.__name__
        # Register config hooks
        self.config.get_item("edit_mode").event += self.update_edit_mode
        self.config.get_item("syntax_theme").event += self.update_style
        self.config.get_item("color_scheme").event += self.update_style
        self.config.get_item("log_level").event += lambda x: setup_logs(self.config)
        self.config.get_item("log_file").event += lambda x: setup_logs(self.config)
        self.config.get_item("log_config").event += lambda x: setup_logs(self.config)
        self.config.get_item("color_depth").event += lambda x: setattr(
            self, "_color_depth", _COLOR_DEPTHS[x.value]
        )
        # Set up the color palette
        self.color_palette = ColorPalette()
        self.color_palette.add_color("fg", "#ffffff", "default")
        self.color_palette.add_color("bg", "#000000", "default")
        # Set up a write position to limit mouse events to a particular region
        self.mouse_limits: WritePosition | None = None
        self.mouse_position = Point(0, 0)

    @property
    def title(self) -> str:
        """The application's title."""
        return self._title

    @title.setter
    def title(self, value: str) -> None:
        """Set the terminal title."""
        self._title = value
        if self.set_title():
            self.output.set_title(value)

    def pause_rendering(self) -> None:
        """Block rendering, but allows input to be processed.

        The first line prevents the display being drawn, and the second line means
        the key processor continues to process keys. We need this as we need to
        wait for the results of terminal queries which come in as key events.

        This is used to prevent flicker when we update the styles based on terminal
        feedback.
        """
        self._is_running = False
        self.renderer._waiting_for_cpr_futures.append(asyncio.Future())

    def resume_rendering(self) -> None:
        """Reume rendering the app."""
        self._is_running = True
        if futures := self.renderer._waiting_for_cpr_futures:
            futures.pop()

    def pre_run(self, app: Application | None = None) -> None:
        """Call during the 'pre-run' stage of application loading."""
        # Load key bindings
        self.load_key_bindings()
        # Determine what color depth to use
        self._color_depth = _COLOR_DEPTHS.get(
            self.config.color_depth, self.term_info.depth_of_color.value
        )
        # Set the application's style, and update it when the terminal responds
        self.update_style()
        self.term_info.colors.event += self.update_style
        # Pause rendering while we load the layout
        self.pause_rendering()
        # Load completions menu. This must be done after the app is initialized, because
        # :py:func:`get_app` is needed to access the config
        self.menus["completions"] = Float(
            content=Shadow(CompletionsMenu()),
            xcursor=True,
            ycursor=True,
        )

        def terminal_ready() -> None:
            """Command here depend on the result of terminal queries."""
            # Open any files we need to
            self.open_files()
            # Load the layout
            # We delay this until we have terminal responses to allow terminal graphics
            # support to be detected first
            self.layout = Layout(self.load_container(), self.focused_element)
            # Run any additional steps
            self.post_load()
            # Resume rendering
            self.resume_rendering()
            # Request cursor position
            self._request_absolute_cursor_position()
            # Sending a repaint trigger
            self.invalidate()
            # Flag that the app is loaded
            self.loaded = True

        if self.input.closed:
            # If we do not have an interactive input, just get on with loading the app:
            # don't send terminal queries, as we will not get responses
            terminal_ready()
        else:
            # Otherwise, we query the terminal and wait asynchronously to give it
            # a chance to respond

            async def await_terminal_feedback() -> None:
                try:
                    # Send queries to the terminal if supported
                    if self.input.__class__.__name__ == "Vt100Input":
                        self.term_info.send_all()
                        # Give the terminal a chance to respond
                        await asyncio.sleep(0.1)
                    # Complete loading the application
                    terminal_ready()
                except Exception as exception:
                    # Log exceptions, as this runs in the event loop and, exceptions may
                    # get hidden from the user
                    log.critical(
                        "An error occurred while trying to load the application",
                        exc_info=True,
                    )
                    self.exit(exception=exception)

            # Waits until the event loop is ready
            self.create_background_task(await_terminal_feedback())

    @classmethod
    def load_input(cls) -> Input:
        """Create the input for this application to use.

        Ensures the TUI app always tries to run in a TTY.

        Returns:
            A prompt-toolkit input instance

        """
        input_ = create_input(always_prefer_tty=True)

        if stdin := getattr(input_, "stdin", None):
            if not stdin.isatty():
                from euporie.core.io import IgnoredInput

                input_ = IgnoredInput()

        # Use a custom vt100 parser to allow querying the terminal
        if parser := getattr(input_, "vt100_parser", None):
            setattr(  # noqa B010
                input_, "vt100_parser", Vt100Parser(parser.feed_key_callback)
            )

        return input_

    @classmethod
    def load_output(cls) -> Output:
        """Create the output for this application to use.

        Ensures the TUI app always tries to run in a TTY.

        Returns:
            A prompt-toolkit output instance

        """
        output = create_output(always_prefer_tty=True)

        if isinstance(output, PtkVt100_Output):
            output = Vt100_Output(
                stdout=output.stdout,
                get_size=output._get_size,
                term=output.term,
                default_color_depth=output.default_color_depth,
                enable_bell=output.enable_bell,
            )

        return output

    def post_load(self) -> None:
        """Allow subclasses to define additional loading steps."""
        # Call extra callables
        for cb in self.post_load_callables:
            cb()

    def load_key_bindings(self) -> None:
        """Load the application's key bindings."""
        from euporie.core.key_binding.bindings.basic import load_basic_bindings
        from euporie.core.key_binding.bindings.micro import load_micro_bindings
        from euporie.core.key_binding.bindings.mouse import load_mouse_bindings

        self._default_bindings = merge_key_bindings(
            [
                # Make sure that the above key bindings are only active if the
                # currently focused control is a `BufferControl`. For other controls, we
                # don't want these key bindings to intervene. (This would break "ptterm"
                # for instance, which handles 'Keys.Any' in the user control itself.)
                ConditionalKeyBindings(
                    merge_key_bindings(
                        [
                            # Load basic bindings.
                            load_ptk_basic_bindings(),
                            load_basic_bindings(),
                            # Load micro bindings
                            load_micro_bindings(config=self.config),
                            # Load emacs bindings.
                            load_emacs_bindings(),
                            load_emacs_search_bindings(),
                            load_emacs_shift_selection_bindings(),
                            # Load Vi bindings.
                            load_vi_bindings(),
                            load_vi_search_bindings(),
                        ]
                    ),
                    buffer_has_focus,
                ),
                # Active, even when no buffer has been focused.
                load_ptk_mouse_bindings(),
                load_cpr_bindings(),
                # Load extra mouse bindings
                load_mouse_bindings(),
                # Load terminal query response key bindings
                # load_command_bindings("terminal"),
            ]
        )
        self.key_bindings = load_registered_bindings(
            *self.bindings_to_load, config=self.config
        )

    def _on_resize(self) -> None:
        """Query the terminal dimensions on a resize event."""
        self.term_info.pixel_dimensions.send()
        super()._on_resize()

    @classmethod
    def launch(cls) -> None:
        """Launch the app."""
        # Load default logging
        setup_logs()
        # Load the app's configuration
        cls.config.load(cls)
        # Configure the logs
        setup_logs(cls.config)
        # Warn about unrecognised configuration items
        cls.config.warn()
        # Run the application
        with create_app_session(input=cls.load_input(), output=cls.load_output()):
            # Create an instance of the app and run it

            original_sigterm = signal.getsignal(signal.SIGTERM)
            original_sigint = signal.getsignal(signal.SIGINT)

            app = cls()

            signal.signal(signal.SIGTERM, app.cleanup)
            signal.signal(signal.SIGINT, app.cleanup)

            result = app.run()

            signal.signal(signal.SIGTERM, original_sigterm)
            signal.signal(signal.SIGINT, original_sigint)

            return result

    def cleanup(self, signum: int, frame: FrameType | None) -> None:
        """Restore the state of the terminal on unexpected exit."""
        log.critical("Unexpected exit signal, restoring terminal")
        output = self.output
        self.exit()
        # Reset terminal state
        # output.quit_alternate_screen()
        # output.disable_mouse_support()
        output.reset_cursor_key_mode()
        output.enable_autowrap()
        # output.disable_bracketed_paste()
        output.clear_title()
        # output.reset_cursor_shape()
        output.show_cursor()
        output.reset_attributes()
        self.renderer.reset()
        # Exit the main thread
        sys.exit(1)

    @classmethod
    async def interact(cls, ssh_session: PromptToolkitSSHSession) -> None:
        """Run the app asynchronously for the hub SSH server."""
        await cls().run_async()

    def load_container(self) -> FloatContainer:
        """Load the root container for this application.

        Returns:
            The root container for this app

        """
        return FloatContainer(
            content=self.layout.container or Window(),
            floats=cast("list[Float]", self.floats),
        )

    def get_file_tabs(self, path: Path) -> list[type[Tab]]:
        """Return the tab to use for a file path."""
        from euporie.core.tabs.base import Tab

        path_mime = get_mime(path) or "text/plain"
        log.debug("File %s has mime type: %s", path, path_mime)

        tab_options = set()
        for tab_cls in Tab._registry:
            for mime_type in tab_cls.mime_types:
                if PurePath(path_mime).match(mime_type):
                    tab_options.add(tab_cls)
            if path.suffix in tab_cls.file_extensions:
                tab_options.add(tab_cls)

        return sorted(tab_options, key=lambda x: x.weight, reverse=True)

    def get_file_tab(self, path: Path) -> type[Tab] | None:
        """Return the tab to use for a file path."""
        if tabs := self.get_file_tabs(path):
            return tabs[0]
        return None

    def open_file(
        self, path: Path, read_only: bool = False, tab_class: type[Tab] | None = None
    ) -> None:
        """Create a tab for a file.

        Args:
            path: The file path of the notebook file to open
            read_only: If true, the file should be opened read_only
            tab_class: The tab type to use to open the file

        """
        ppath = parse_path(path)
        log.info(f"Opening file {path}")
        for tab in self.tabs:
            if ppath == getattr(tab, "path", "") and (
                tab_class is None or isinstance(tab, tab_class)
            ):
                log.info(f"File {path} already open, activating")
                self.layout.focus(tab)
                break
        else:
            if tab_class is None:
                tab_class = self.get_file_tab(path)
            if tab_class is None:
                log.error("Unable to display file %s", path)
            else:
                tab = tab_class(self, ppath)
                self.tabs.append(tab)
                # Ensure the opened tab is focused at app start
                self.focused_element = tab
                # Ensure the newly opened tab is selected
                self.tab_idx = len(self.tabs) - 1

    def open_files(self) -> None:
        """Open the files defined in the configuration."""
        for file in self.config.files:
            self.open_file(file)

    @property
    def tab(self) -> Tab | None:
        """Return the currently selected tab container object."""
        if self.tabs:
            # Detect if focused tab has changed
            # Find index of selected child
            for i, tab in enumerate(self.tabs):
                if self.render_counter > 0 and self.layout.has_focus(tab):
                    self._tab_idx = i
                    break
            self._tab_idx = max(0, min(self._tab_idx, len(self.tabs) - 1))
            return self.tabs[self._tab_idx]
        else:
            return None

    @property
    def tab_idx(self) -> int:
        """Get the current tab index."""
        return self._tab_idx

    @tab_idx.setter
    def tab_idx(self, value: int) -> None:
        """Set the current tab by index."""
        self._tab_idx = value % (len(self.tabs) or 1)
        if self.tabs:
            try:
                self.layout.focus(self.tabs[self._tab_idx])
            except ValueError:
                pass

    def focus_tab(self, tab: Tab) -> None:
        """Make a tab visible and focuses it."""
        self.tab_idx = self.tabs.index(tab)

    def cleanup_closed_tab(self, tab: Tab) -> None:
        """Remove a tab container from the current instance of the app.

        Args:
            tab: The closed instance of the tab container

        """
        # Remove tab
        self.tabs.remove(tab)
        # Update body container to reflect new tab list
        # assert isinstance(self.body_container.body, HSplit)
        # self.body_container.body.children[0] = VSplit(self.tabs)
        # Focus another tab if one exists
        if self.tab:
            self.layout.focus(self.tab)
        # If a tab is not open, the status bar is not shown, so focus the logo, so
        # pressing tab focuses the menu
        else:
            try:
                self.layout.focus_next()
            except ValueError:
                pass

    def close_tab(self, tab: Tab | None = None) -> None:
        """Close a notebook tab.

        Args:
            tab: The instance of the tab to close. If `None`, the currently
                selected tab will be closed.

        """
        if tab is None:
            tab = self.tab
        if tab is not None:
            tab.close(cb=partial(self.cleanup_closed_tab, tab))

    def get_edit_mode(self) -> EditingMode:
        """Return the editing mode enum defined in the configuration."""
        from euporie.core.key_binding.bindings.micro import EditingMode

        return {
            "micro": EditingMode.MICRO,  # type: ignore
            "vi": EditingMode.VI,
            "emacs": EditingMode.EMACS,
        }.get(
            str(self.config.edit_mode), EditingMode.MICRO  # type: ignore
        )

    def update_edit_mode(self, setting: Setting | None = None) -> None:
        """Set the keybindings for editing mode."""
        self.editing_mode = self.get_edit_mode()
        log.debug("Editing mode set to: %s", self.editing_mode)

    def create_merged_style(self) -> BaseStyle:
        """Generate a new merged style for the application.

        Using a dynamic style has serious performance issues, so instead we update
        the style on the renderer directly when it changes in `self.update_style`

        Returns:
            Return a combined style to use for the application

        """
        # Get foreground and background colors based on the configured colour scheme
        theme_colors: dict[str, dict[str, str]] = {
            "default": {},
            "light": {"fg": "#202020", "bg": "#F0F0F0"},
            "dark": {"fg": "#F0F0F0", "bg": "#202020"},
            "white": {"fg": "#000000", "bg": "#FFFFFF"},
            "black": {"fg": "#FFFFFF", "bg": "#000000"},
            # TODO - use config.custom_colors
            "custom": {
                "fg": self.config.custom_foreground_color,
                "bg": self.config.custom_background_color,
            },
        }
        base_colors: dict[str, str] = {
            **DEFAULT_COLORS,
            **self.term_info.colors.value,
            **theme_colors.get(self.config.color_scheme, theme_colors["default"]),
        }

        # Build a color palette from the fg/bg colors
        self.color_palette = ColorPalette()
        for name, color in base_colors.items():
            self.color_palette.add_color(
                name,
                color or theme_colors["default"][name],
                "default" if name in ("fg", "bg") else name,
            )
        # Add accent color
        self.color_palette.add_color(
            "hl", base_colors.get(self.config.accent_color, self.config.accent_color)
        )

        # Build app style
        app_style = build_style(self.color_palette)

        # Apply style transformations based on the configured color scheme
        self.style_transformation = merge_style_transformations(
            [
                ConditionalStyleTransformation(
                    SetDefaultColorStyleTransformation(
                        fg=base_colors["fg"], bg=base_colors["bg"]
                    ),
                    self.config.color_scheme != "default",
                ),
                ConditionalStyleTransformation(
                    SwapLightAndDarkStyleTransformation(),
                    self.config.color_scheme == "inverse",
                ),
            ]
        )

        return merge_styles(
            [
                style_from_pygments_cls(get_style_by_name(self.config.syntax_theme)),
                Style(MIME_STYLE),
                Style(HTML_STYLE),
                Style(LOG_STYLE),
                Style(IPYWIDGET_STYLE),
                app_style,
            ]
        )

    def update_style(
        self,
        query: TerminalQuery | Setting | None = None,
    ) -> None:
        """Update the application's style when the syntax theme is changed."""
        self.renderer.style = self.create_merged_style()

    def refresh(self) -> None:
        """Reet all tabs."""
        for tab in self.tabs:
            to_container(tab).reset()

    def _create_merged_style(
        self, include_default_pygments_style: Filter | None = None
    ) -> BaseStyle:
        """Block default style loading."""
        return DummyStyle()

    def draw(self, render_as_done: bool = True) -> None:
        """Draw the app without focus, leaving the cursor below the drawn output."""
        # Hide ephemeral containers
        self._redrawing = True
        # Ensure nothing in the layout has focus
        self.layout._stack.append(Window())
        # Re-draw the app
        self._redraw(render_as_done=render_as_done)
        # Remove the focus block
        self.layout._stack.pop()
        # Show ephemeral containers
        self._redrawing = False
        # Ensure the renderer knows where the cursor is
        self._request_absolute_cursor_position()

    def _handle_exception(
        self, loop: AbstractEventLoop, context: dict[str, Any]
    ) -> None:
        exception = context.get("exception")
        # Log observed exceptions to the log
        log.exception("An unhandled exception occurred", exc_info=exception)

    # async def cancel_and_wait_for_background_tasks(self) -> "None":
    # """Cancel background tasks, ignoring exceptions."""
    # # try:
    # # await super().cancel_and_wait_for_background_tasks()
    # # except ValueError as e:
    # # for task in self._background_tasks:
    # # print(task)
    # # raise e
    #
    # for task in self._background_tasks:
    # print(task)
    # print(task.get_loop())
    # await asyncio.wait([task])

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd()
    def _quit() -> None:
        """Quit euporie."""
        get_app().exit()

    @staticmethod
    @add_cmd(
        name="close-tab",
        filter=tab_has_focus,
        menu_title="Close File",
    )
    def _close_tab() -> None:
        """Close the current tab."""
        get_app().close_tab()

    @staticmethod
    @add_cmd(
        filter=tab_has_focus,
    )
    def _next_tab() -> None:
        """Switch to the next tab."""
        get_app().tab_idx += 1

    @staticmethod
    @add_cmd(
        filter=tab_has_focus,
    )
    def _previous_tab() -> None:
        """Switch to the previous tab."""
        get_app().tab_idx -= 1

    @staticmethod
    @add_cmd(
        filter=~buffer_has_focus,
    )
    def _focus_next() -> None:
        """Focus the next control."""
        get_app().layout.focus_next()

    @staticmethod
    @add_cmd(
        filter=~buffer_has_focus,
    )
    def _focus_previous() -> None:
        """Focus the previous control."""
        get_app().layout.focus_previous()

    @staticmethod
    @add_cmd()
    def _clear_screen() -> None:
        """Clear the screen."""
        get_app().renderer.clear()

    # ################################### Settings ####################################w

    add_setting(
        name="files",
        default=[],
        flags=["files"],
        nargs="*",
        type_=UPath,
        help_="List of file names to open",
        schema={
            "type": "array",
            "items": {
                "description": "File path",
                "type": "string",
            },
        },
        description="""
            A list of file paths to open when euporie is launched.
        """,
    )

    add_setting(
        name="edit_mode",
        flags=["--edit-mode"],
        type_=str,
        choices=["micro", "emacs", "vi"],
        title="Editor key bindings",
        help_="Key-binding mode for text editing",
        default="micro",
        description="""
            Key binding style to use when editing cells.
        """,
    )

    add_setting(
        name="tab_size",
        flags=["--tab-size"],
        type_=int,
        help_="Spaces per indentation level",
        default=4,
        schema={
            "minimum": 1,
        },
        description="""
            The number of spaces to use per indentation level. Should be set to 4.
        """,
    )

    add_setting(
        name="terminal_polling_interval",
        flags=["--terminal-polling-interval"],
        type_=int,
        help_="Time between terminal colour queries",
        default=0,
        schema={
            "min": 0,
        },
        description="""
            Determine how frequently the terminal should be polled for changes to the
            background / foreground colours. Set to zero to disable terminal polling.
        """,
    )

    add_setting(
        name="autoformat",
        flags=["--autoformat"],
        type_=bool,
        help_="Automatically re-format code cells when run",
        default=False,
        description="""
            Whether to automatically reformat code cells before they are run.
        """,
    )

    add_setting(
        name="format_black",
        flags=["--format-black"],
        type_=bool,
        help_="Use black when re-formatting code cells",
        default=True,
        description="""
            Whether to use :py:mod:`black` when reformatting code cells.
        """,
    )

    add_setting(
        name="format_isort",
        flags=["--format-isort"],
        type_=bool,
        help_="Use isort when re-formatting code cells",
        default=True,
        description="""
            Whether to use :py:mod:`isort` when reformatting code cells.
        """,
    )

    add_setting(
        name="format_ssort",
        flags=["--format-ssort"],
        type_=bool,
        help_="Use ssort when re-formatting code cells",
        default=True,
        description="""
            Whether to use :py:mod:`ssort` when reformatting code cells.
        """,
    )

    add_setting(
        name="syntax_theme",
        flags=["--syntax-theme"],
        type_=str,
        help_="Syntax highlighting theme",
        default="euporie",
        schema={
            # Do not want to print all theme names in `--help` screen as it looks messy
            # so we only add them in the scheme, not as setting choices
            "enum": list(pygments_styles.keys()),
        },
        description="""
            The name of the pygments style to use for syntax highlighting.
        """,
    )

    add_setting(
        name="color_depth",
        flags=["--color-depth"],
        type_=int,
        choices=[1, 4, 8, 24],
        default=None,
        help_="The color depth to use",
        description="""
            The number of bits to use to represent colors displayable on the screen.
            If set to None, the supported color depth of the terminal will be detected
            automatically.
        """,
    )

    add_setting(
        name="tmux_graphics",
        flags=["--tmux-graphics"],
        type_=bool,
        help_="Enable terminal graphics in tmux (experimental)",
        default=False,
        hidden=~in_tmux,
        description="""
            If set, terminal graphics will be used if :program:`tmux` is running by
            performing terminal escape sequence pass-through. You must restart euporie
            for this to take effect.

            You will also need to ensure that ``allow-passthrough`` is set to ``on`` in
            your :program:`tmux` configuration.

            .. warning::

               Terminal graphics in :program:`tmux` is experimental, and is not
               guaranteed to work. Use at your own risk!
        """,
    )

    add_setting(
        name="color_scheme",
        flags=["--color-scheme"],
        type_=str,
        choices=["default", "inverse", "light", "dark", "black", "white", "custom"],
        help_="The color scheme to use",
        default="default",
        description="""
            The color scheme to use: `auto` means euporie will try to use your
            terminal's color scheme, `light` means black text on a white background,
            and `dark` means white text on a black background.
        """,
    )

    add_setting(
        name="custom_background_color",
        flags=["--custom-background-color", "--custom-bg-color", "--bg"],
        type_=str,
        help_='Background color for "Custom" color theme',
        default="",
        schema={
            "maxLength": 7,
        },
        description="""
            The hex code of the color to use for the background in the "Custom" color
            scheme.
        """,
    )

    add_setting(
        name="custom_foreground_color",
        flags=["--custom-foreground-color", "--custom-fg-color", "--fg"],
        type_=str,
        help_='Foreground color for "Custom" color theme',
        default="",
        schema={
            "maxLength": 7,
        },
        description="""
            The hex code of the color to use for the foreground in the "Custom" color
            scheme.
        """,
    )

    add_setting(
        name="accent_color",
        flags=["--accent-color"],
        type_=str,
        help_="Accent color to use in the app",
        default="ansiblue",
        description="""
            The hex code of a color to use for the accent color in the application.
        """,
    )

    add_setting(
        name="key_bindings",
        flags=["--key-bindings"],
        type_=json.loads,
        help_="Additional key binding definitions",
        default={},
        description="""
            A mapping of component names to mappings of command name to key-binding lists.
    """,
        schema={
            "type": "object",
        },
    )

    add_setting(
        name="graphics",
        flags=["--graphics"],
        choices=["none", "sixel", "kitty", "iterm"],
        type_=str,
        default=None,
        help_="The preferred graphics protocol",
        description="""
            The graphics protocol to use, if supported by the terminal.
            If set to "none", terminal graphics will not be used.
    """,
    )

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.core.app.BaseApp": {
                "quit": "c-q",
                "close-tab": "c-w",
                "next-tab": "c-pagedown",
                "previous-tab": "c-pageup",
                "focus-next": "s-tab",
                "focus-previous": "tab",
                "clear-screen": "c-l",
            }
        }
    )



"""Register additional key escape sequences."""

from itertools import combinations

from aenum import extend_enum
from prompt_toolkit.input.ansi_escape_sequences import ANSI_SEQUENCES
from prompt_toolkit.keys import Keys

# Map key-modifier to key-name & emacs style shortcut prefix
_modifiers = {
    # 0b00000000: ("", ""),
    0b00000001: ("Shift", "s"),
    0b00000010: ("Alt", "A"),
    0b00000100: ("Control", "c"),
    # 0b00001000: ("Super", "S"),
    # 0b00010000: ("Meta", "M"),
    # 0b00100000: ("Hyper", "H"),
    0b01000000: ("", ""),  # ScollLock - pass-through
    0b10000000: ("", ""),  # NumLock - pass-through
}

_key_nos = {
    9: "tab",
    13: "enter",
    27: "escape",
    32: "space",
    **{i: chr(i) for i in range(126, 32, -1)},
    127: "backspace",
}


# Kitty style legacy function keys
_kitty_functional_codes = {
    "Insert": (2, "~"),
    "Delete": (3, "~"),
    "PageUp": (5, "~"),
    "PageDown": (6, "~"),
    "Up": (1, "A"),
    "Down": (1, "B"),
    "Right": (1, "C"),
    "Left": (1, "D"),
    "Home": (1, "H"),
    "End": (1, "F"),
    "PrintScreen": (57361, "u"),
    "Pause": (57362, "u"),
    "Menu": (57363, "u"),
    "F1": (1, "P"),
    "F2": (1, "Q"),
    "F3": (13, "~"),
    "F4": (1, "S"),
    "F5": (15, "~"),
    "F6": (17, "~"),
    "F7": (18, "~"),
    "F8": (19, "~"),
    "F9": (20, "~"),
    "F10": (21, "~"),
    "F11": (23, "~"),
    "F12": (24, "~"),
    # "Menu": (29, "~"),
}

# Add Alt-key shortcuts
for key in _key_nos.values():
    key_var = f"Alt{key.title()}"
    if not hasattr(Keys, key_var):
        extend_enum(Keys, key_var, f"A-{key}")
    key_enum = getattr(Keys, key_var)
    # Alias escape + key for Alt-key
    ANSI_SEQUENCES[f"\x1b{key}"] = key_enum  # type: ignore

# Add CSI-u escape key
ANSI_SEQUENCES["\x1b[27u"] = Keys.Escape

# Add various CSI-u style keys
for n in range(len(_modifiers)):
    for mod_combo in combinations(_modifiers.items(), n):
        mod_name = "".join(name for _bit, (name, _short) in mod_combo[::-1])
        mod_short = "-".join(short for _bit, (_name, short) in mod_combo[::-1] if _name)
        mod_no = 1 + sum(bit for bit, _ in mod_combo)

        if mod_name:
            for i, key in _key_nos.items():
                key_var = f"{mod_name}{key.title()}"
                if not hasattr(Keys, key_var):
                    extend_enum(Keys, key_var, f"{mod_short}-{key}")
                key_enum = getattr(Keys, key_var)
                # CSI-u style
                ANSI_SEQUENCES[f"\x1b[{i};{mod_no}u"] = key_enum  # type: ignore
                # xterm style
                ANSI_SEQUENCES[f"\x1b[27;{mod_no};{i}~"] = key_enum  # type: ignore

        for key, (number, suffix) in _kitty_functional_codes.items():
            key_var = f"{mod_name}{key}"
            if not hasattr(Keys, key_var):
                mod_str = f"{mod_short}-" if mod_short else ""
                extend_enum(Keys, key_var, f"{mod_str}{key.lower()}")
            mod_str = str(mod_no)
            num_str = str(number)
            if mod_no == 1:
                mod_str = ""
                if number == 1:
                    num_str = ""
            seq = "\x1b[" + (";".join(x for x in [num_str, mod_str] if x)) + suffix
            ANSI_SEQUENCES[seq] = getattr(Keys, key_var)  # type: ignore
